from ...kawa_base_e2e_test import KawaBaseTest
import pandas as pd
import uuid
import tempfile
from ....client.kawa_client import KawaClient as K


class TestComputationWithTextFilters(KawaBaseTest):

    @classmethod
    def setUpClass(cls):
        unique_id = 'resource_{}'.format(uuid.uuid4())
        print(tempfile.gettempdir())
        KawaBaseTest.setUpClass()
        cls.df = cls._df_for_tests()
        cls.sheet_name = unique_id
        loader = cls.kawa.new_data_loader(df=cls.df.copy(deep=True), datasource_name=unique_id)
        loader.create_datasource()
        loader.load_data(reset_before_insert=True, create_sheet=True)

    def test_without_filters(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .select('text', 'float')
                       .compute())

        # then
        self.assertListEqual(list(computed_df['text']), ['this', 'is', 'some', 'test', 'data', ''])
        self.assertListEqual(list(computed_df['float'])[:-1], [1.1, 2.2, 3.3, 4.4, 5.5])
        self.assertTrue(pd.isna(list(computed_df['float'])[-1]))

    def test_or_composition(self):
        self._check_filter(K.col('text').contains('i').contains('a'), ['this', 'is', 'data'])
        self._check_filter(K.col('float').gt(4.4).lt(2.05), ['this', 'data'])

    def test_or_and_exclude_composition(self):
        self._check_filter(K.col('text').contains('i').contains('a').exclude(), ['some', 'test', ''])

    def test_and_composition(self):
        # when
        df = (self.kawa
              .sheet(sheet_name=self.sheet_name)
              .filter(K.col('text').contains('i'))
              .filter(K.col('text').contains('h'))
              .select('text')
              .compute())

        # then
        self.assertListEqual(list(df['text']), ['this'])

    def _check_filter(self, f, expected):
        df = (self.kawa
              .sheet(sheet_name=self.sheet_name)
              .filter(f)
              .select('text')
              .compute())
        self.assertListEqual(list(df['text']), expected)

    @staticmethod
    def _df_for_tests():
        data = {
            'float': [1.1, 2.2, 3.3, 4.4, 5.5, None],
            'text': ['this', 'is', 'some', 'test', 'data', '']
        }
        return pd.DataFrame(data)



class KawaLayoutBuilder:

    def __init__(self, kawa_client, sheet_id, view_name, view_type):
        self._k = kawa_client
        self._cmd = self._k.commands
        self._sheet_id = str(sheet_id)
        sheet = self._k.entities.sheets().get_entity(self._sheet_id)
        self._created_layout_id = None
        self._primary_ds_id = sheet['primaryDataSourceId']
        self._filters = []
        self._view_name = view_name
        self._view_type = view_type
        self._top = {
            "sortDirection": "ASCENDING",
            "ofDepth": -1,
            "columnId": "NO_TOP",
            "count": -1,
            "perDepth": -1,
            "applicable": True,
            "aggregationMethod": "NOOP"
        }

    def set_top_dimension(self, dimension, measure, aggregation='SUM', how_many=5):
        self._top = {
            "aggregationMethod": aggregation,
            "columnId": self._column_id(measure),
            "count": how_many,
            "sortDirection": "DESCENDING",
            "groupByColumnId": self._column_id(dimension)
        }
        return self

    def add_text_filter(self, filter_column, allowed_values):
        self._filters.append({
            "allowedValues": allowed_values,
            "mode": "ValuesList",
            "applyTo": self._column_id(filter_column),
            "enabled": True,
            "identifierType": "COLUMN_ID",
            "filterType": "TEXT_FILTER",
            "initial": False,
            "depth": 0,
            "applicable": True,
            "aggregationMethod": "FIRST"
        })
        return self

    def add_rolling_range_filter(self, filter_column, range_from=-1, range_to=0, unit='DAY'):
        self._filters.append({
            "selectionSize": max(range_from, range_to),
            "selectionOffset": min(range_from, range_to),
            "temporalUnit": unit,
            "dateSelectionMethod": "ROLLING_SPAN",
            "mode": "RANGE",
            "applyTo": self._column_id(filter_column),
            "enabled": True,
            "identifierType": "COLUMN_ID",
            "filterType": "DATE_FILTER",
            "initial": False,
            "depth": 0,
            "applicable": True,
            "aggregationMethod": "MAX"
        })
        return self

    def add_numeric_range_filter(self, filter_column, range_from=None, range_to=None):
        rules = []
        if range_from is None and range_to is None:
            return
        if range_from is not None:
            rules.append({"checkAgainst": range_from, "checkMethod": "GREATER_THAN_OR_EQUALS"})
        if range_to is not None:
            rules.append({"checkAgainst": range_to, "checkMethod": "LESSER_THAN_OR_EQUALS"})

        self._filters.append({
            "operatorBetweenRules": "AND",
            "mode": "ValuesRange",
            "rules": rules,
            "applyTo": self._column_id(filter_column),
            "enabled": True,
            "identifierType": "COLUMN_ID",
            "filterType": "NUMBER_FILTER",
            "initial": False,
            "depth": 0,
            "applicable": True,
            "aggregationMethod": "SUM"
        })
        return self

    def initialize(self, standalone=True):
        print("Creating {} {}".format(self._view_type, self._view_name))
        self._created_layout_id = self._run(
            command_name='CreateLayout',
            command_parameters={
                'layoutType': self._view_type,
                'sheetId': self._sheet_id,
                'status': 'ACTIVE',
                'createLayoutWithoutFields': self._view_type == 'CHART',
                'standalone': standalone
            }
        ).get('id')
        self._run(
            command_name='RenameEntity',
            command_parameters={
                'id': self._created_layout_id,
                'displayInformation': {
                    'displayName': self._view_name,
                    'description': ''
                },
                "entityType": "layout"
            }
        )
        return self

    def _run(self, command_name, command_parameters):
        return self._cmd.run_command(
            command_name=command_name,
            command_parameters=command_parameters
        )

    def _column_id(self, column_name):
        return "{}⟶{}".format(self._primary_ds_id, column_name)

    def _apply_all_filter(self):
        self._run(
            command_name='ReplaceFiltering',
            command_parameters={
                "layoutId": self._created_layout_id,
                "filtering": {
                    "top": self._top,
                    "accessFilters": [],
                    "columnFilters": self._filters,
                    "attributeFilters": [],
                    "fieldFilters": []
                }
            }
        )


class KawaGridBuilder(KawaLayoutBuilder):

    def __init__(self, kawa_client, sheet_id, grid_name):
        super().__init__(kawa_client=kawa_client, sheet_id=sheet_id, view_name=grid_name, view_type='GRID')
        self._k = kawa_client
        self._cmd = self._k.commands
        self._sheet_id = str(sheet_id)
        sheet = self._k.entities.sheets().get_entity(self._sheet_id)
        self._primary_ds_id = sheet['primaryDataSourceId']
        self._grid_name = grid_name
        self._created_layout_id = None
        self._filters = []

    def set_grouping(self, columns):
        grouping_items = [{"fieldId": self._column_id(c)} for c in columns]
        self._run(
            command_name='ReplaceGrouping',
            command_parameters={
                "layoutId": self._created_layout_id,
                "rowGrouping": {
                    "groupingItems": grouping_items
                }
            }
        )
        return self

    def set_sorting(self, column, direction='DESCENDING'):
        self._run(command_name='ReplaceSorting',
                  command_parameters={
                      "layoutId": self._created_layout_id,
                      "sorting": {
                          "fieldSortingList": [
                              {
                                  "fieldId": self._column_id(column),
                                  "sortDirection": direction
                              }
                          ]
                      }
                  })
        return self

    def hide_fields(self, columns):
        self._run(command_name='ReplaceHiddenFields',
                  command_parameters={
                      "layoutId": self._created_layout_id,
                      "hiddenFieldIds": [self._column_id(c) for c in columns]
                  })
        return self

    def set_aggregation(self, column, aggregation_method):
        self._run(command_name='ReplaceFieldAggregation',
                  command_parameters={
                      "layoutId": self._created_layout_id,
                      "fieldId": self._column_id(column),
                      "aggregationMethod": aggregation_method
                  })
        return self

    def finalize(self):
        self._apply_all_filter()
        return self._created_layout_id


class KawaChartBuilder(KawaLayoutBuilder):

    def __init__(self, kawa_client, sheet_id, chart_name, chart_type):
        super().__init__(kawa_client=kawa_client, sheet_id=sheet_id, view_name=chart_name, view_type='CHART')
        self._chart_name = chart_name
        self._chart_type = chart_type
        self._series_id = None
        self._chart_type = chart_type

    def finalize(self,
                 legend_position='NONE',
                 bar_stacking_mode=0,
                 show_labels=True,
                 num_labels=100,
                 label_rotation=45,
                 line_area=False,
                 color=0):
        self._apply_all_filter()
        self._set_display_configuration(legend_position=legend_position,
                                        bar_stacking_mode=bar_stacking_mode,
                                        show_labels=show_labels,
                                        num_labels=num_labels,
                                        label_rotation=label_rotation,
                                        line_area=line_area,
                                        color=color)
        return self._created_layout_id

    def sort_series(self, direction='DESCENDING'):
        self._run(command_name='ReplaceSorting',
                  command_parameters={
                      "layoutId": self._created_layout_id,
                      "sorting": {
                          "fieldSortingList": [
                              {
                                  "fieldId": self._series_id,
                                  "sortDirection": direction
                              }
                          ]
                      }
                  })
        return self

    def add_comparison(self, comparison_type='PREVIOUS'):
        self._run(
            command_name='AddChartComparison',
            command_parameters={
                "layoutId": self._created_layout_id,
                "comparisons": [
                    {
                        "comparisonParameters": {
                            "fieldId": self._series_id
                        },
                        "comparisonType": comparison_type,
                        "enabled": True
                    }
                ]
            }
        )
        return self

    def add_grouping(self, column_id, name=None):
        self._run(
            command_name='AddChartGrouping',
            command_parameters={
                "layoutId": self._created_layout_id,
                "columnId": column_id,
                "insertFirst": False,
                "displayInformation": {
                    "displayName": name if name else column_id,
                    "description": ""
                }
            }
        )
        return self

    def set_sampling(self, column_id, name=None):
        self._run(
            command_name='UpdateChartGrouping',
            command_parameters={
                "layoutId": self._created_layout_id,
                "columnId": column_id,
                "insertFirst": False,
                "displayInformation": {
                    "displayName": name if name else column_id,
                    "description": ""
                }
            }
        )
        return self

    def add_series(self, column_id, name, aggregation='COUNT', unit=None):
        chart = self._run(
            command_name='AddChartSeries',
            command_parameters={
                "layoutId": self._created_layout_id,
                "columnId": column_id,
                "displayInformation": {
                    "displayName": "Measure",
                    "description": ""
                },
                "seriesType": self._chart_type
            }
        )
        series_id = chart.get('seriesIds')[0]['fieldId']
        self._run(
            command_name='RenameField',
            command_parameters={
                "layoutId": self._created_layout_id,
                "fieldId": series_id,
                "fieldName": name,
            }
        )
        self._run(
            command_name='ReplaceFieldAggregation',
            command_parameters={
                "layoutId": self._created_layout_id,
                "aggregationMethod": aggregation,
                "fieldId": series_id,
            }
        )
        if unit:
            self._run(
                command_name='ReplaceFieldFormatters',
                command_parameters={
                    "layoutId": self._created_layout_id,
                    "formatters": {
                        "decimal": {
                            "precision": {
                                "params": {
                                    "fractionDigits": 2,
                                    "numberLocale": "local"
                                }
                            },
                            "unit": {
                                "params": {
                                    "unit": "auto"
                                }
                            },
                            "postfix": {
                                "params": {
                                    "postfix": unit
                                }
                            },
                            "negative": {
                                "params": {
                                    "negativeType": "minus"
                                }
                            }
                        },
                        "integer": {
                            "precision": {
                                "params": {
                                    "fractionDigits": 0,
                                    "numberLocale": "local"
                                }
                            },
                            "unit": {
                                "params": {
                                    "unit": "auto"
                                }
                            },
                            "postfix": {
                                "params": {
                                    "postfix": unit
                                }
                            },
                            "negative": {
                                "params": {
                                    "negativeType": "minus"
                                }
                            }
                        }
                    },
                    "fieldId": self._series_id,
                }
            )
        return self

    def _set_display_configuration(self,
                                   legend_position='NONE',
                                   bar_stacking_mode=0,
                                   show_labels=True,
                                   num_labels=100,
                                   label_rotation=45,
                                   line_area=False,
                                   color=0):
        self._run(
            command_name='ReplaceChartSeriesType',
            command_parameters={
                "layoutId": self._created_layout_id,
                "seriesTypes": {
                    self._series_id: self._chart_type,
                }
            }
        )
        self._run(
            command_name='ReplaceChartDisplayConfiguration',
            command_parameters={
                "layoutId": self._created_layout_id,
                "chartDisplayConfiguration": {
                    "chartType": self._chart_type,
                    "series": [
                        {
                            "id": "s_59ab",
                            "isVisible": True,
                            "type": self._chart_type,
                            "colorIndexInPalette": color,
                            "fieldId": self._series_id,
                            "label": show_labels
                        }
                    ],
                    "map": {
                        "seriesToYAxis": {
                            "s_59ab": "a_b237"
                        },
                        "yAxisToContainer": {
                            "a_4483": "g_cef6",
                            "a_ce86": "g_cef6",
                            "a_b237": "g_cef6",
                            "a_3912": "g_cef6"
                        }
                    },
                    "yAxis": [
                        {
                            "type": "value",
                            "id": "a_b237"
                        },
                        {
                            "type": "value",
                            "id": "a_3912"
                        },
                        {
                            "type": "value",
                            "id": "a_4483"
                        },
                        {
                            "type": "value",
                            "id": "a_ce86"
                        }
                    ],
                    "comparisonsConfig": {},
                    "containers": [
                        {
                            "id": "g_cef6"
                        },
                        {
                            "id": "g_db28"
                        },
                        {
                            "id": "g_2472"
                        },
                        {
                            "id": "g_99eb"
                        }
                    ],
                    "gridLines": True,
                    "multigrid": False,
                    "lineAreaStyle": line_area,
                    "labelItemsNumber": num_labels,
                    "labelItemRotation": label_rotation,
                    "stacking": bar_stacking_mode,
                    "showDataZoom": True,
                    "smoothLine": False,
                    "showYAxisLabel": True,
                    "showPoints": True,
                    "legend": [
                        {
                            "positionMode": legend_position
                        },
                        {
                            "positionMode": "NONE"
                        }
                    ],
                    "formatters": {},
                    "lineWidth": 1,
                    "isMultiSeriesMode": True
                }}
        )



from .app_widget import SimpleChart, Table, ScatterChart, TextWidget, IndicatorChart, WidgetFactory
from uuid import uuid4
from .app_synchronizer import Synchronizer

from .application_building_utils import load_dashboard, info, error

NUM_COLUMNS = 32


class ApplicationPage:

    def __init__(self, kawa, reporter, application, data_model, name: str):
        self._name = name
        self._data_model = data_model
        self._application = application
        self._sections = []
        self._dashboard_id = None
        self._k = kawa
        self._reporter = reporter

    @property
    def name(self):
        return self._name

    @property
    def tag(self):
        return f'{self._application.tag}|{self._name}'

    @property
    def dashboard_id(self):
        return self._dashboard_id

    @property
    def widgets(self):
        widget_collector = []
        for section in self._sections:
            for column in section.columns:
                widget_collector += column.widgets()
        return widget_collector

    def create_section(self, title, num_columns=1):

        if num_columns == 0:
            raise Exception(f'Please define at least one column for section {title}')
        if num_columns > 8:
            raise Exception(f'Please define at most eight columns for section {title}')

        new_section = ApplicationPageSection(
            kawa=self._k,
            reporter=self._reporter,
            page=self,
            title=title,
            data_model=self._data_model,
            num_columns=num_columns,
        )
        self._sections.append(new_section)
        columns = new_section.columns
        return columns[0] if num_columns == 1 else columns

    def sync(self, application_id):

        if not self._sections:
            raise Exception('You have to define at least one section by calling the create_section() function')

        self._dashboard_id = ApplicationPage._Synchronizer(
            kawa=self._k,
            page=self,
            application_id=application_id,
        ).sync()

        for section in self._sections:
            section.sync()

        self.set_layout()

        self._reporter.report(
            object_type='Dashboard',
            name=self._name,
        )

    def set_layout(self):
        current_position = 0

        # Propagate layout down to the widgets
        for section in self._sections:
            section.set_position(current_position)
            current_position = section.set_layout()

        dashboard_layout = [w.position() for w in self.widgets]
        self._k.commands.run_command('ReplaceDashboardLayout', {
            "dashboardId": str(self._dashboard_id),
            "dashboardLayout": dashboard_layout,
        })

    class _Synchronizer(Synchronizer):
        def __init__(self, kawa, page, application_id):
            super().__init__(
                kawa=kawa,
                icon='🎨',
                entity_description=f'Page "{page.name}"',
                entity_tag=page.tag,
            )
            self._page = page
            self._application_id = application_id
            self._new_dashboard_id = None

        def _load_state(self):
            return self._k.entities.dashboards().find_entity_by_tag(tag=self._tag)

        def _raise_if_state_invalid(self):
            ...

        def _should_create(self):
            return not self._state

        def _create_new_entity(self):
            component_id = str(uuid4())
            updated_application = self._k.commands.run_command(
                command_name='addNewDashboardToApplication',
                command_parameters={
                    "applicationId": str(self._application_id),
                    "componentId": component_id,
                    "layoutType": "ONE_PAGE",
                    "tag": self._tag,
                }
            )['application']

            components = updated_application.get('components', [])
            created_component = [c for c in components if c['componentId'] == component_id]
            if not created_component:
                raise Exception(f'The page {self._name} could not be created in the application')

            self._new_dashboard_id = created_component[0]['componentConfiguration']['entityId']
            self._k.commands.run_command(
                command_name='renameEntity',
                command_parameters={
                    "displayInformation": {
                        "displayName": self._page.name,
                        "description": ""
                    },
                    "entityType": "dashboard",
                    "id": str(self._new_dashboard_id)
                }
            )

        def _update_entity(self):
            dashboard_id = self._state['id']
            widgets = self._state['widgets']
            if widgets:
                widget_ids = [w['widgetId'] for w in widgets]
                self._k.commands.run_command(
                    command_name='deleteWidget',
                    command_parameters={
                        "dashboardId": str(dashboard_id),
                        "widgetIds": widget_ids,
                    }
                )

        def _build_new_state(self):
            return self._state['id'] if self._state else self._new_dashboard_id


class ApplicationPageSection:

    def __init__(self, kawa, reporter, page, data_model, title, num_columns):
        self._k = kawa
        self._reporter = reporter
        self._title = title
        self._page = page
        self._data_model = data_model
        self._columns = [
            ApplicationPageColumn(
                kawa=kawa,
                reporter=self._reporter,
                section=self,
                data_model=data_model,
                width=NUM_COLUMNS // num_columns,
                position=i * NUM_COLUMNS // num_columns
            ) for i in range(num_columns)
        ]

        # Position is the y coordinate of the top of the section
        self._position = None

    def set_position(self, y):
        self._position = y

    @property
    def position(self):
        return self._position

    @property
    def dashboard_id(self):
        return self._page.dashboard_id

    @property
    def columns(self):
        return self._columns

    def sync(self):
        info(f'🖼️ Sync a new section {self._title} with {len(self._columns)} columns')
        for column in self._columns:
            column.sync()

    def set_layout(self):
        max_y = 0
        for column in self._columns:
            y = column.set_layout()
            if y > max_y:
                max_y = y

        # Will be used to position the next section
        return max_y


class ApplicationPageColumn:

    def __init__(self, kawa, reporter, position, width, section, data_model):
        self._k = kawa
        self._reporter = reporter
        self._section = section
        self._data_model = data_model
        self._widgets = []

        # Position of the column on the x-axis and width of the column
        self._position = position
        self._width = width

        # This will be used to track the position of the current widget
        # And incremented each time we add a new one
        self._current_y = 0

        # To build widgets
        self._widget_factory = WidgetFactory(
            kawa=kawa,
            dashboard_id_supplier=lambda: self._section.dashboard_id,
            default_sheet_id_supplier=lambda: self._data_model.sheet_id,
        )

    def widgets(self):
        return self._widgets

    def set_layout(self):
        self._current_y = self._section.position
        for widget in self._widgets:
            widget.set_position(
                x=self._position,
                y=self._current_y,
                width=self._width,
            )

            self._current_y += widget.compute_height(self._width)

        return self._current_y

    def indicator_chart(self, **kwargs):
        chart = self._widget_factory.indicator_chart(**kwargs)
        self._widgets.append(chart)

    def scatter_chart(self, **kwargs):
        chart = self._widget_factory.scatter_chart(**kwargs)
        self._widgets.append(chart)

    def bar_chart(self, **kwargs):
        chart = self._widget_factory.bar_chart(**kwargs)
        self._widgets.append(chart)

    def line_chart(self, **kwargs):
        chart = self._widget_factory.line_chart(**kwargs)
        self._widgets.append(chart)

    def pie_chart(self, **kwargs):
        chart = self._widget_factory.pie_chart(**kwargs)
        self._widgets.append(chart)

    def boxplot(self, **kwargs):
        chart = self._widget_factory.boxplot(**kwargs)
        self._widgets.append(chart)

    def table(self, **kwargs):
        chart = self._widget_factory.table(**kwargs)
        self._widgets.append(chart)

    def text_widget(self, html):
        widget = TextWidget(
            kawa=self._k,
            dashboard_id_supplier=lambda: self._section.dashboard_id,
            content=html,
        )
        self._widgets.append(widget)

    def sync(self):
        for widget in self._widgets:
            try:
                widget.sync()
            except Exception as e:
                error(f'⚠️Error on sync for widget {widget.title}: {e}')
                raise e



from abc import ABC, abstractmethod
import time


class Synchronizer(ABC):

    def __init__(self, kawa, icon, entity_description, entity_tag=None):
        self._k = kawa
        self._icon = icon
        self._entity_description = entity_description
        self._tag = entity_tag
        self._state = None
        self._step_id = 0

    def sync(self):
        start = time.time()
        self._step('Start Synchronization')

        self._step('Loading State')
        self._state = self._load_state()

        self._step('Checking State')
        self._raise_if_state_invalid()

        if self._should_create():
            self._step('Creating a new object')
            self._create_new_entity()
        else:
            self._step('Updating existing object')
            self._update_entity()

        new_state = self._build_new_state()
        end = time.time()
        self._step(f'Synchronization done in {end - start:.1f}s')

        return new_state

    def _step(self, step):
        self._step_id += 1
        if self._step_id == 1:
            tag = f' (🏷️{self._tag})' if self._tag else ''
            print(f'{self._icon} {self._entity_description}{tag}:')
        print(f'   {step}')

    @abstractmethod
    def _load_state(self):
        ...

    @abstractmethod
    def _raise_if_state_invalid(self):
        ...

    @abstractmethod
    def _should_create(self):
        ...

    @abstractmethod
    def _create_new_entity(self):
        ...

    @abstractmethod
    def _update_entity(self):
        ...

    @abstractmethod
    def _build_new_state(self):
        ...



from .app_synchronizer import Synchronizer


class Metric:

    def __init__(self, kawa, reporter, description, name, sql, prompt, ai_mode=False):
        if (prompt is None) == (sql is None):
            raise Exception('Define one (And only one): query or prompt')

        self._k = kawa
        self._reporter = reporter
        self._name = name
        self._description = description
        self._sql = sql
        self._prompt = prompt
        self._ai_mode = ai_mode

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @property
    def sql(self):
        return self._sql

    @property
    def prompt(self):
        return self._prompt

    def to_ascii(self):
        return f'"{self._name}"={self._sql or self._prompt}'.replace("SELECT", "")

    def sync(self, sheet):
        Metric._Synchronizer(
            kawa=self._k,
            ai_mode=self._ai_mode,
            metric=self,
            sheet=sheet
        ).sync()

        self._reporter.report(
            object_type='Metric',
            name=self._name,
        )

    class _Synchronizer(Synchronizer):

        def __init__(self, kawa, ai_mode, metric, sheet):
            super().__init__(
                kawa=kawa,
                icon='📐',
                entity_description=f'Metric "{metric.name}"',
            )
            self._metric = metric
            self._sheet = sheet
            self._ai_mode = ai_mode

        def _load_state(self):
            if self._ai_mode:
                # State is not required in AI Mode, we will always be adding new adhoc columns
                return {}
            else:
                return {c['displayInformation']['displayName']: c for c in self._sheet['computedColumns'] if
                        c['columnNature'] == 'COMPUTATION'}

        def _raise_if_state_invalid(self):
            ...

        def _should_create(self) -> bool:
            if self._ai_mode:
                # Always create in AI Mode, because we create adHoc columns
                return True
            else:
                existing_formulas = self._state
                return self._metric.name not in existing_formulas

        def _create_new_entity(self):
            metric_definition = self._prepare_metric_definition()
            sheet_id = self._sheet['id']
            self._k.commands.run_command('addComputedColumnToSheet', {
                'sheetId': str(sheet_id),
                'displayInformation': {
                    'displayName': self._metric.name,
                    'description': self._metric.description,
                },
                'addInDefaultLayout': False,
                "adHoc": self._ai_mode,
                **metric_definition,
            })

        def _update_entity(self):
            metric_definition = self._prepare_metric_definition()
            existing_formulas = self._state
            column_id = existing_formulas[self._metric.name]['columnId']
            sheet_id = self._sheet['id']
            self._k.commands.run_command('updateColumnXmlSyntacticTree', {
                'sheetId': str(sheet_id),
                'columnId': str(column_id),
                **metric_definition,
            })

        def _build_new_state(self):
            ...

        def _prepare_metric_definition(self):
            metric = self._metric
            sheet_id = self._sheet['id']
            if metric.prompt:
                generated_xml = self._k.post(
                    url=f'{self._k.kawa_api_url}/gen-ai/generate-xml-formula',
                    data={
                        'prompt': metric.prompt,
                        'sheetId': str(sheet_id),
                    }
                )['generatedXml']
                return {'xmlSyntacticTree': generated_xml}
            else:
                return {'sql': metric.sql}



import pandas as pd
from datetime import date
from ...client.kawa_client import KawaClient
from ...client.kawa_decorators import kawa_tool


def kawa():
    kawa = KawaClient(kawa_api_url='http://localhost:4200')
    kawa.set_api_key(api_key_file='/Users/emmanuel/doc/local-pristine/.key')
    kawa.set_active_workspace_id(workspace_id='75')
    return kawa


def profit_section(page):
    col1 = page.create_section('Profit', 1)
    col1.bar_chart(
        title='Profit and Net Profit per State',
        x='State',
        y=('Profit', 'Net Profit'),
        aggregation=('SUM', 'SUM'),
        color_offset=1,
    )


def raw_data_section(page):
    col = page.create_section('Table', 1)
    col.table('Store Data')


def kpis_section(page):
    col1, col2, col3, col4 = page.create_section('KPIs', 4)
    col1.indicator_chart('Total Profit ($)', 'Profit')
    col2.indicator_chart('Total Sales ($)', 'Sales')
    col3.indicator_chart('Total Quantity', 'Quantity')
    col4.indicator_chart('Avg Discount ($)', 'Discount', 'AVERAGE')


def trends_section(page):
    col1 = page.create_section('Trends', 1)
    col1.line_chart(
        title='Evolution of Profit and Sales ($)',
        x='Order Date',
        y=('Profit', 'Sales'),
        aggregation=('SUM', 'SUM'),
        time_sampling='YEAR_AND_MONTH',
        show_values=False,
        area=True,
        fill_in_temporal_gaps=True,
    )


def breakdowns_section(page):
    col1, col2 = page.create_section('Breakdowns', 2)
    col1.bar_chart(
        title='Shipping delay per Ship Mode (days)',
        x='Ship Mode',
        y='Shipping Delay',
        aggregation='AVERAGE',
        color='Segment',
        show_totals=True,
    )
    col2.bar_chart(
        title='Profit per Sub-Category and Ship Mode ($)',
        x='Sub-Category',
        y='Profit',
        color='Ship Mode',
        show_totals=True,
    )
    col1.bar_chart(
        title='Profit per Year and Segment ($)',
        x='Order Date',
        y='Profit',
        color='Segment',
        time_sampling='YEAR',
        show_values=True,
        show_totals=True,
    )
    col2.bar_chart(
        title='Profit per Year and Ship Mode ($)',
        x='Order Date',
        y='Profit',
        color='Ship Mode',
        time_sampling='YEAR',
        show_values=True,
        show_totals=True,
    )


def pies_section(page):
    col1, col2, col3, col4 = page.create_section('Pies', 4)
    col1.pie_chart(
        title='Quantity per Ship Mode',
        labels='Ship Mode',
        values='Quantity',
        show_values=True,
        show_labels=True,
    )
    col2.pie_chart(
        title='Quantity per Year',
        labels='Order Date',
        values='Quantity',
        show_values=True,
        show_labels=True,
        time_sampling='YEAR',
    )
    col3.pie_chart(
        title='Quantity per Segment',
        labels='Segment',
        values='Quantity',
        show_values=True,
        show_labels=True,
        doughnut=True,
    )
    col4.pie_chart(
        title='Quantity per Category',
        labels='Category',
        values='Quantity',
        show_values=True,
        show_labels=True,
        doughnut=True,
    )


def scatters_section(page):
    col1, col2 = page.create_section('Scatters', 2)

    col1.scatter_chart(
        title='Profit vs. Quantity per City',
        granularity='City',
        x='Profit',
        y='Quantity',
        color='Order ID',
        aggregation_color='COUNT'
    )

    col2.scatter_chart(
        title='Profit vs. Quantity per Customer',
        granularity='Customer Name',
        x='Profit',
        y='Quantity',
        color='Order ID',
        aggregation_color='COUNT'
    )

    col1.scatter_chart(
        title='Profit vs. Quantity per Date',
        granularity='Order Date',
        x='Profit',
        y='Quantity',
        color='Order ID',
        aggregation_color='COUNT'
    )

    col2.scatter_chart(
        title='Profit vs. Quantity per Product',
        granularity='Product Name',
        x='Profit',
        y='Quantity',
        color='Order ID',
        aggregation_color='COUNT'
    )


@kawa_tool(outputs={
    'Row ID': float,
    'Order ID': str,
    'Order Date': date,
    'Ship Date': date,
    'Ship Mode': str,
    'Customer ID': str,
    'Customer Name': str,
    'Segment': str,
    'Country': str,
    'City': str,
    'State': str,
    'Postal Code': str,
    'Region': str,
    'Product ID': str,
    'Category': str,
    'Sub-Category': str,
    'Product Name': str,
    'Sales': float,
    'Quantity': float,
    'Discount': float,
    'Profit': float,
})
def load_store_data():
    url = "https://learn.kawa.ai/readme-assets/store.csv"
    df = pd.read_csv(
        url,
        parse_dates=['Ship Date', 'Order Date'],
        date_format='%Y-%m-%d',
    )
    df['Postal Code'] = df['Postal Code'].astype('string')
    return df


@kawa_tool(outputs={
    'event id': float,
    'event date': date,
    'guest name': str,
    'event type': str,
    'event cost': float,
    'event state': str,
})
def load_events_data():
    url = "https://learn.kawa.ai/readme-assets/events.csv"
    df = pd.read_csv(
        url,
        parse_dates=['event date'],
        date_format='%Y-%m-%d',
    )
    return df


app = kawa().app(
    'Superstore App II',
    sidebar_color="#040506",
    theme='classic1'
)

app.create_ai_agent(
    name='BinaryBard',
    instructions='''
    You are an expert in data analytics. 
    Always answer with as much precision as possible.
    ''',
    color="#040506",
)

store_dataset = app.create_dataset('Store', load_store_data)
event_dataset = app.create_dataset('event', load_events_data)

model = app.create_model(store_dataset)

events_relationship = model.create_relationship(
    name='Event per Customer',
    dataset=event_dataset,
    link={'State': 'event state'},
)
events_relationship.add_column(
    name='event cost',
    aggregation='SUM',
    new_column_name='Cost per State',
)
model.create_metric(
    name='Net Profit',
    formula='SUM("Profit") - SUM("Cost per State")'
)
model.create_metric(
    name='Shipping Delay',
    prompt='How many days between ship date and order date'
)


# Add filters to the application
@kawa_tool(parameters={'state': {'type': str, 'binding': 'State'}})
def value_loader(state):
    print(f'Figuring out the list of Segments for {state}')
    return ['Consumer','Corporate', 'Home Office']


app.create_text_filter('State')
app.create_text_filter('Segment', value_loader=value_loader)

# Build Order Analytics page
orders_page = app.create_page('Orders Analytics')
kpis_section(orders_page)
scatters_section(orders_page)
trends_section(orders_page)
pies_section(orders_page)
breakdowns_section(orders_page)
raw_data_section(orders_page)
profit_section(orders_page)

app.publish()



import ast

# Replace this string with the path to your Python file
file_path = '/Users/emmanuel/dev/kywy/src/kywy/client/application_builder_test.py'


def extract_globals(file_path):
    with open(file_path, 'r') as f:
        tree = ast.parse(f.read())

    globals = []
    for node in tree.body:  # only top-level (global scope)
        if isinstance(node, ast.Assign):
            # handle multiple targets (e.g., a = b = 3)
            for target in node.targets:
                if isinstance(target, ast.Name):
                    globals.append(target.id)
        elif isinstance(node, ast.AnnAssign):
            # annotated assignment (e.g., x: int = 5)
            target = node.target
            if isinstance(target, ast.Name):
                globals.append(target.id)
    return globals


if __name__ == "__main__":
    vars = extract_globals(file_path)
    print("Global variables:", vars)



import inspect
import ast

from typing import Callable
from .application_building_utils import IMPORTS, info
from .app_synchronizer import Synchronizer


class DataSet:

    def __init__(self, kawa, reporter, name: str, application, generator: Callable = None, datasource_id=None):
        self._k = kawa
        self._reporter = reporter
        self._name = name
        self._generator = generator
        self._datasource_id = datasource_id
        self._application = application
        self._sheet = None

    @property
    def name(self):
        return self._name

    @property
    def datasource_id(self):
        return self._datasource_id

    @property
    def sheet_id(self):
        return self._sheet['id'] if self._sheet else None

    @property
    def sheet(self):
        return self._sheet

    @property
    def tag(self):
        return f'{self._application.tag}|{self._name}'

    def sync(self):

        self._datasource_id = DataSet._DataSynchronizer(
            kawa=self._k,
            dataset=self,
        ).sync()

        self._sheet = DataSet._SheetSynchronizer(
            kawa=self._k,
            dataset=self,
        ).sync()

        self._reporter.report(
            object_type='DataSet',
            name=self._name,
        )

    def extract_script_source_code(self):
        return '{}\n\n{}'.format(
            '\n'.join(IMPORTS),
            inspect.getsource(self._generator)
        )

    def extract_indicator_types_from_script(self):
        source_code = self.extract_script_source_code()
        return self._extract_outputs(source_code)

    @staticmethod
    def _extract_outputs(source_code):
        tree = ast.parse(source_code)
        for node in ast.walk(tree):
            if isinstance(node, ast.FunctionDef):
                for deco in node.decorator_list:
                    if isinstance(deco, ast.Call) and getattr(deco.func, 'id', '') == 'kawa_tool':
                        for keyword in deco.keywords:
                            if keyword.arg == 'outputs':
                                outputs_node = keyword.value
                                outputs_src = ast.unparse(outputs_node)
                                safe_globals = {
                                    'str': 'text',
                                    'float': 'decimal',
                                    'date': 'date',
                                    'datetime': 'date_time',
                                    'bool': 'boolean',
                                }
                                return eval(outputs_src, safe_globals, {})

        return None

    class _DataSynchronizer(Synchronizer):

        def __init__(self, kawa, dataset):
            super().__init__(
                kawa=kawa,
                icon='💿',
                entity_description=f'Dataset "{dataset.name}"',
                entity_tag=dataset.tag
            )
            self._dataset = dataset
            self._new_datasource_id = None

        def _load_state(self):
            existing_datasource = self._k.entities.datasources().find_entity_by_tag(tag=self._tag)
            existing_script = self._k.entities.scripts().find_entity_by_tag(tag=self._tag)
            existing_etl_pipeline = self._k.entities.etl_pipelines().find_entity_by_tag(tag=self._tag)
            return {
                'datasource': existing_datasource,
                'script': existing_script,
                'etl': existing_etl_pipeline
            }

        def _raise_if_state_invalid(self):
            # The state is valid only if nothing is defined or if everything is defined
            num_defined = self._num_defined_entities()
            if num_defined != 0 and num_defined != len(self._state):
                raise Exception('The state for the DataSet is not consistent')

        def _should_create(self):
            return self._num_defined_entities() == 0

        def _create_new_entity(self):
            self._script_id = self._k.commands.run_command(
                command_name='createScript',
                command_parameters={
                    'name': self._dataset.name,
                    'tag': self._tag,
                    'content': self._dataset.extract_script_source_code()
                }
            )['id']

            indicator_types = self._dataset.extract_indicator_types_from_script()
            if not indicator_types:
                raise Exception('No indicators were found in the the underlying script')

            indicators = [self._build_indicator_from_id_and_type(indicator_id, indicator_type)
                          for indicator_id, indicator_type in indicator_types.items()]

            created_datasource_id = self._k.commands.run_command(
                command_name='createEtlAndDatasource',
                command_parameters=self._build_datasource_definition(indicators)
            )['dataSource']['id']

            self._new_datasource_id = created_datasource_id

        def _update_entity(self):
            self._update_script_if_necessary()
            self._update_datasource_if_necessary()

        def _build_new_state(self):
            datasource_state = self._state['datasource']
            return datasource_state['id'] if datasource_state else self._new_datasource_id

        def _update_script_if_necessary(self):
            existing_script = self._state['script']
            existing_etl_pipeline = self._state['etl']
            script_file_id = existing_script['sourceControlToolConfiguration']['toolName']
            existing_script_tmp_file = self._k.download_file_as_id(script_file_id)
            with open(existing_script_tmp_file, "r", encoding="utf-8") as file:
                existing_content = file.read()

            new_content = self._dataset.extract_script_source_code()

            if new_content != existing_content:
                self._k.commands.run_command('replaceScript', {
                    'content': new_content,
                    'name': existing_script['displayInformation']['displayName'],
                    'scriptId': str(existing_script['id'])
                })
                self._k.commands.run_command('runEtl', {
                    "etlPipelineId": str(existing_etl_pipeline['id']),
                    "fullRefresh": False,
                })

        def _update_datasource_if_necessary(self):
            existing_datasource = self._state['datasource']
            new_indicators = self._dataset.extract_indicator_types_from_script()
            existing_indicators = {
                i['displayInformation']['displayName']: i['type']
                for i in existing_datasource['indicators']
                if i['indicatorId'] != 'record_id'
            }

            # Indicators that are in the existing datasource but not in the new generator function
            # (Nothing to do, it just means that some indicators will no longer have a value)
            missing_indicator_names = []

            # Indicators that are in the new generator function but not in the existing data source
            # (We will need to add them)
            indicators_to_add = {}

            # Incompatible indicator: Indicators whose type changed
            # Records the name and the type that the indicator should have
            incompatible_indicators = {}

            # Looking at the new indicators, we are able to detect the ones that were not here initially
            # and the ones whose type has changed
            for new_indicator_name, new_indicator_type in new_indicators.items():
                if new_indicator_name not in existing_indicators:
                    indicators_to_add[new_indicator_name] = new_indicator_type
                else:
                    existing_indicator_type = existing_indicators[new_indicator_name]
                    if existing_indicator_type != new_indicator_type:
                        incompatible_indicators[new_indicator_name] = existing_indicator_type

            # Looking at the existing indicators, we are able to detect what is missing in the new generator function
            for existing_indicator_name in existing_indicators.keys():
                if existing_indicator_name not in new_indicators:
                    missing_indicator_names.append(existing_indicator_name)

            if incompatible_indicators:
                raise Exception(
                    'These indicators should have the following types, please check it: {}'
                    .format(incompatible_indicators))
            elif indicators_to_add:
                info('- Datasource: Adding {} indicators'.format(len(indicators_to_add)))
                self._k.commands.run_command(
                    command_name='addIndicatorsToDataSource',
                    command_parameters={
                        'dataSourceId': str(existing_datasource['id']),
                        'newIndicators': [
                            {
                                'displayInformation': {'displayName': name},
                                'includedInDefaultLayout': True,
                                'indicatorId': name,
                                'type': kawa_type,
                            }
                            for name, kawa_type in indicators_to_add.items()
                        ],
                    }
                )

        def _num_defined_entities(self):
            return len([val for val in self._state.values() if val])

        def _build_datasource_definition(self, indicators):
            return {
                "doNotCreateSheet": True,
                "isMapping": False,
                "displayInformation": {
                    "displayName": self._dataset.name,
                    "description": "",
                    "extraInformation": {
                        "immutableTag": self._tag
                    }
                },
                "loadingAdapterName": 'CLICKHOUSE',
                "loadingMode": "RESET_BEFORE_INSERT",
                "extractionAdapterName": "PYTHON_SCRIPT",
                "extractionAdapterConfiguration": {
                    "scriptId": str(self._script_id),
                    "scriptParametersValues": []
                },
                "indicators": indicators,
                "jobTrigger": {
                    "enabled": False,
                    "scheduleType": "INTERVAL",
                    "onlyOnBusinessDays": None,
                    "interval": 3,
                    "timeUnit": "HOURS",
                    "timeZone": None,
                    "initialDelay": 0
                },
                "rowMapperConfigList": [],
                "defaultGlobalPolicy": "ALLOW_ALL",
                "createAssociatedTimeSeriesDataSource": False,
                "needsQueryCostCheck": False
            }

        @staticmethod
        def _build_indicator_from_id_and_type(indicator_id, indicator_type):
            return {
                "elementDataModel": None,
                "displayInformation": {
                    "displayName": indicator_id,
                    "description": ""
                },
                "possibleValues": [],
                "indicatorId": indicator_id,
                "indicatorKind": "TABLE",
                "indicatorStatus": "ACTIVE",
                "includedInDefaultLayout": True,
                "type": indicator_type
            }

    class _SheetSynchronizer(Synchronizer):

        def __init__(self, kawa, dataset):
            super().__init__(
                kawa=kawa,
                icon='📙',
                entity_description=f'Sheet "{dataset.name}"',
                entity_tag=dataset.tag,
            )
            self._dataset = dataset
            self._new_sheet = None

        def _load_state(self):
            existing_datasource = self._k.entities.datasources().find_entity_by_tag(tag=self._tag)
            existing_sheet = self._k.entities.sheets().find_entity_by_tag(tag=self._tag)
            return {
                'datasource': existing_datasource,
                'sheet': existing_sheet,
            }

        def _raise_if_state_invalid(self):
            sheet_is_defined = bool(self._state['sheet'])
            data_is_defined = bool(self._state['datasource'])
            if sheet_is_defined and not data_is_defined:
                raise Exception('The Sheet is defined but not the datasource')

        def _should_create(self):
            return not self._state['sheet']

        def _create_new_entity(self):
            datasource_id = self._state['datasource']['id']
            self._new_sheet = self._k.commands.run_command(
                command_name='createSheet',
                command_parameters={
                    "createDefaultLayout": True,
                    "sheet": {
                        "displayInformation": {
                            "displayName": self._dataset.name,
                            "description": "",
                            "extraInformation": {
                                "immutableTag": self._tag
                            }
                        },
                        "dataSourceLinks": [{"targetDataSourceId": str(datasource_id)}],
                        "shared": False,
                        "generalAccess": "RESTRICTED"
                    }
                }
            )

        def _update_entity(self):
            self._new_sheet = self._k.commands.run_command(
                command_name='resetSheet',
                command_parameters={
                    "sheetId": str(self._state['sheet']['id'])
                }
            )

        def _build_new_state(self):
            return self._k.entities.sheets().find_entity_by_tag(tag=self._tag)



from datetime import datetime

import json

from .app_model import DataModel
from .app_reporter import Reporter
from .report_builder import ReportBuilder
from .app_widget import WidgetFactory


class ResearchBuilder:

    def __init__(self, kawa_client, name):
        self._k = kawa_client
        self._name = name
        self._models = []
        self._results = []
        self._charts = []

        self._reporter = Reporter(name=name)
        self._widget_factory = WidgetFactory(
            kawa=kawa_client
        )

    def publish_main_model(self, main_model):
        for model in self._models:
            model.sync()

        relationships = self._analyse_relationships()
        metrics = self._analyse_metrics()

        return json.dumps(
            obj={
                'relationships': relationships,
                'metrics': metrics,
                'mainSheetId': main_model.sheet_id,
            },
            indent=4
        )

    def publish_results(self):
        created_grids = []
        created_charts = []

        created_chart_collector = {}
        for chart in self._charts:
            created_layout = chart.sync()
            sheet_id = created_layout['sheet_id']
            if created_layout['sheet_id'] not in created_chart_collector:
                created_chart_collector[sheet_id] = []
            created_chart_collector[sheet_id].append(created_layout['layout_id'])

        for sheet_id, view_ids in created_chart_collector.items():
            created_charts.append({
                'sheetId': sheet_id,
                'viewIds': view_ids
            })

        relationships = self._analyse_relationships()

        for model in self._models:
            view_ids = model.create_views()
            created_grids.append({
                'sheetId': str(model.sheet_id),
                'viewIds': view_ids
            })

        result = {
            'grids': [v for v in created_grids if v['viewIds']],
            'charts': [v for v in created_charts if v['viewIds']],
            'data': self._results,
            'relationships': relationships,
        }

        return json.dumps(result)

    def register_result(self, description, df, from_select=True):
        self._results.append({
            'description': description,
            'fromSelect': from_select,
            'df': df.to_csv(),
        })

    def register_model(self, model_id, **kwargs):
        sheet = self._k.entities.sheets().get_entity_by_id(model_id)
        if not sheet:
            raise Exception(f'Sheet with id={model_id} not found')
        model = DataModel(
            kawa=self._k,
            reporter=self._reporter,
            name=sheet['displayInformation']['displayName'],
            sheet=sheet,
            ai_mode=True, # Research are meant for AI
        )
        self._models.append(model)
        return model

    # Charts
    def bar_chart(self, **kwargs):
        self._register_chart(self._widget_factory.bar_chart, **kwargs)

    def line_chart(self, **kwargs):
        self._register_chart(self._widget_factory.line_chart, **kwargs)

    def indicator_chart(self, **kwargs):
        self._register_chart(self._widget_factory.indicator_chart, **kwargs)

    def boxplot(self, **kwargs):
        self._register_chart(self._widget_factory.boxplot, **kwargs)

    def scatter_chart(self, **kwargs):
        self._register_chart(self._widget_factory.scatter_chart, **kwargs)

    def pie_chart(self, **kwargs):
        self._register_chart(self._widget_factory.pie_chart, **kwargs)

    def _register_chart(self, chart_generator, **kwargs):
        kwargs['sheet_id'] = kwargs['model'].sheet_id
        del kwargs['model']
        chart = chart_generator(**kwargs)
        self._charts.append(chart)

    def _analyse_relationships(self):
        relationships = []
        for model in self._models:
            ascii_rep = [rel.to_ascii() for rel in model.relationships]
            if ascii_rep:
                relationships.append({
                    'sheetName': model.name,
                    'asciiRepresentations': ascii_rep,
                })
        return relationships

    def _analyse_metrics(self):
        metrics = []
        for model in self._models:
            ascii_rep = [m.to_ascii() for m in model.metrics]
            if ascii_rep:
                metrics.append({
                    'sheetName': model.name,
                    'asciiRepresentations': ascii_rep,
                })
        return metrics



from .app_variable import Variable
from .app_metric import Metric
from .app_relationship import Relationship


class DataModel:

    def __init__(self, kawa, reporter, name, dataset=None, sheet=None, ai_mode=False):
        self._dataset = dataset
        self._name = name
        self._k = kawa
        self._ai_mode = ai_mode
        self._reporter = reporter
        self._relationships = []
        self._metrics = []
        self._variables = []
        self._selects = []
        self._created_grid_ids = []
        self._sheet = sheet

    @property
    def sheet_id(self):
        return self._sheet['id'] if self._sheet else self._dataset.sheet_id

    @property
    def sheet(self):
        return self._sheet if self._sheet else self._dataset.sheet

    @property
    def name(self):
        return self._name

    @property
    def relationships(self):
        return self._relationships

    @property
    def metrics(self):
        return self._metrics

    def create_variable(self, name, kawa_type, initial_value):
        variable = Variable(
            kawa=self._k,
            reporter=self._reporter,
            sheet_id_supplier=lambda: self.sheet_id,
            name=name,
            kawa_type=kawa_type,
            initial_value=initial_value
        )
        self._variables.append(variable)

    def create_relationship(self,
                            name,
                            link,
                            description=None,
                            # These two are synonyms
                            dataset=None, origin_model=None,
                            target_model=None,
                            ):
        rel = Relationship(
            kawa=self._k,
            reporter=self._reporter,
            model=self,
            name=name,
            description=description,
            dataset=dataset or origin_model,
            target_sheet=target_model.sheet if target_model else None,
            link=link,
            ai_mode=self._ai_mode,
        )
        self._relationships.append(rel)
        return rel

    def create_metric(self, name, description=None, formula=None, prompt=None, **kwargs):

        sql = None
        if formula:
            normalized = formula.strip().upper()
            if not normalized.startswith("SELECT"):
                sql = f"SELECT {formula}"
            else:
                sql = formula

        self._metrics.append(
            Metric(
                kawa=self._k,
                reporter=self._reporter,
                name=name,
                description=description or sql or prompt,
                sql=sql,
                prompt=prompt,
                ai_mode=self._ai_mode,
            )
        )

    def select(self, *columns_or_column_names):
        select = self._k.sheet(sheet_id=self.sheet_id).select(*columns_or_column_names)
        self._selects.append(select)
        return select

    def create_views(self):
        i = 1
        for select in self._selects:
            created_grid = select.as_grid(
                standalone=False
            )
            self._created_grid_ids.append(created_grid['id'])
            i += 1

        return self._created_grid_ids

    def sync(self):
        for rel in self._relationships:
            rel.sync()

        for var in self._variables:
            var.sync()

        for metric in self._metrics:
            metric.sync(sheet=self.sheet)



from ..kawa_base_e2e_test import KawaBaseTest
import pandas as pd
import uuid
import tempfile
from ...client.kawa_client import KawaClient as K


class TestComputationWithSorting(KawaBaseTest):
    cities_and_countries = pd.DataFrame([
        {'id': 'a', 'country': 'FR', 'city': 'Paris', 'measure': 1},
        {'id': 'b', 'country': 'FR', 'city': 'Lyon', 'measure': 2},
        {'id': 'c', 'country': 'FR', 'city': 'Lille', 'measure': 3},
        {'id': 'd', 'country': 'UK', 'city': 'London', 'measure': 4},
        {'id': 'e', 'country': 'UK', 'city': 'Cardiff', 'measure': 5},
        {'id': 'g', 'country': 'UK', 'city': 'Edinburgh', 'measure': 6},
        {'id': 'h', 'country': 'UK', 'city': 'Belfast', 'measure': 7},
        {'id': 'i', 'country': 'BE', 'city': 'Brussels', 'measure': 8},
        {'id': 'j', 'country': 'BE', 'city': 'Bruges', 'measure': 9},
        {'id': 'k', 'country': 'BE', 'city': 'Namur', 'measure': 10},
    ])

    @classmethod
    def setUpClass(cls):
        unique_id = 'resource_{}'.format(uuid.uuid4())
        print(tempfile.gettempdir())
        KawaBaseTest.setUpClass()
        cls.sheet_name = unique_id

    def test_that_aliases_are_taken_in_account(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=self.cities_and_countries, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        df = (self.kawa.sheet(sheet_name=unique_id)
              .order_by('city', ascending=True)
              .select(K.col('measure').max().alias('max'), K.col('measure').min().alias('min'))
              .group_by('1')
              .compute())

        # then
        self.assertListEqual(list(df.columns), ['grouping(0)', 'max', 'min'])
        self.assertListEqual(list(df['grouping(0)']), ['1'])
        self.assertListEqual(list(df['max']), [10])
        self.assertListEqual(list(df['min']), [1])

    def test_that_aliases_can_be_used_in_sorting(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=self.cities_and_countries, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        df = (self.kawa.sheet(sheet_name=unique_id)
              .order_by('max', ascending=True)
              .select(K.col('measure').max().alias('max'), K.col('measure').min().alias('min'))
              .compute())

        # then
        self.assertListEqual(list(df.columns), ['max', 'min'])
        self.assertListEqual(list(df['max']), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        self.assertListEqual(list(df['min']), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

    def test_that_aliases_can_be_used_in_grouping(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=self.cities_and_countries, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        df = (self.kawa
              .sheet(sheet_name=unique_id)
              .order_by('max', ascending=True)
              .select(
                K.col('measure').max().alias('max'),
                K.col('measure').min().alias('min'),
                K.col('country').alias('nation'),
              )
              .group_by('nation')
              .compute())

        # then
        self.assertListEqual(list(df.columns), ['grouping(0)', 'max', 'min', 'nation'])
        self.assertListEqual(list(df['grouping(0)']), ['FR', 'UK', 'BE'])
        self.assertListEqual(list(df['max']), [3, 7, 10])
        self.assertListEqual(list(df['min']), [1, 4, 8])
        self.assertListEqual(list(df['nation']), ['FR', 'UK', 'BE'])

    def test_that_aliases_can_be_used_in_filter(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=self.cities_and_countries, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        df = (self.kawa
              .sheet(sheet_name=unique_id)
              .order_by('max', ascending=True)
              .select(
                K.col('measure').max().alias('max'),
                K.col('measure').min().alias('min'),
                K.col('country').alias('nation'),
              )
              .filter(K.col('country').eq('FR'))
              .group_by('nation')
              .compute())

        # then
        self.assertListEqual(list(df.columns), ['grouping(0)', 'max', 'min', 'nation'])
        self.assertListEqual(list(df['grouping(0)']), ['FR'])
        self.assertListEqual(list(df['max']), [3])
        self.assertListEqual(list(df['min']), [1])
        self.assertListEqual(list(df['nation']), ['FR'])




from .source_code_utils import extract_sections, extract_between_marker_lines, extract_before_marker_line, \
    extract_between_fence
from anthropic import Anthropic
from openai import OpenAI
import os
import tempfile
import importlib
import uuid
import io
import sys
import traceback
from datetime import datetime

script_dir = os.path.dirname(os.path.abspath(__file__))

PATH_TO_LOG_FILE = '/tmp/co-builder.log'
MODEL = 'claude-sonnet-4-20250514'
MARKERS = [
    '-- DATA SECTION',
    '-- MODEL SECTION',
    '-- DASHBOARD SECTION'
]
SYSTEM_PROMPT_PATHS = {
    'create': f'{script_dir}/prompts/full-prompt.md',
    'edit': f'{script_dir}/prompts/edit-prompt.md',
}


class NotebookChat:

    def __init__(self):
        self._initial_script = None
        self._conversation = []
        self._patches = []
        self._anthropic_client = Anthropic(
            api_key=os.getenv('ANTHROPIC_API_KEY')
        )
        self._open_ai_client = OpenAI(
            api_key=os.getenv('OPENAI_API_KEY')
        )
        self._qwen_client = OpenAI(
            api_key=os.getenv('QWEN_API_KEY'),
            base_url="https://dashscope-intl.aliyuncs.com/compatible-mode/v1",
        )

    @staticmethod
    def log(message):
        with open(PATH_TO_LOG_FILE, "a") as f:
            now = datetime.now()
            current_time = now.strftime("%Y-%m-%d %H:%M:%S")
            f.write(f"[{current_time}] {message}\n")

    def reset(self):
        self._initial_script = None
        self._conversation = []
        self._patches = []

    def execute_code(self):
        merged_code = self._merged_code()
        with tempfile.NamedTemporaryFile('w', suffix='.py', delete=False) as tmp_file:
            extracted_code = extract_between_fence(merged_code)
            tmp_file.write(extracted_code)
            tmp_path = tmp_file.name

        stdout_buffer = io.StringIO()
        original_stdout = sys.stdout
        sys.stdout = StreamDuplicator(original_stdout, stdout_buffer)

        try:
            module_name = f"mod_{uuid.uuid4().hex}"
            spec = importlib.util.spec_from_file_location(module_name, tmp_path)
            mod = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(mod)

        finally:
            os.remove(tmp_path)
            sys.stdout = original_stdout
            captured_output = stdout_buffer.getvalue()

            error_section = extract_between_marker_lines(
                captured_output,
                marker='# ERROR SECTION'
            ).strip()

            if error_section:
                return error_section

    def generate_code(self, prompt):

        is_first_iteration = self._first_iteration()
        action = 'create' if is_first_iteration else 'edit'
        system_prompt = self._build_system_prompt(action)

        self._conversation.append({
            "role": "user",
            "content": prompt
        })

        generated_content = self._generate_content_claude(system_prompt)

        self._conversation.append({
            "role": "assistant",
            "content": generated_content
        })

        if is_first_iteration:
            self._initial_script = generated_content
        else:
            extractions = extract_sections(
                source=generated_content,
                markers=MARKERS,
                patch_name=prompt
            )
            self._patches.append(extractions)

        return generated_content

    def generate_application_specs(self, prompt):
        instructions = self._load_prompt_file('research', 'research')
        tools = [{"type": "web_search"}]
        response = self._generate_content_open_ai(
            client=self._open_ai_client,
            model="gpt-5-nano",
            system_prompt=instructions
        )
        return response

    def _merged_code(self):
        sections = extract_sections(self._initial_script, markers=MARKERS)
        for marker in MARKERS:
            if marker not in sections:
                raise Exception(f'The initial script is missing this section: {marker}')

        for patch in self._patches:
            sections.update(patch)

        script_header = extract_before_marker_line(self._initial_script, MARKERS[0])
        final_script = script_header

        for marker in MARKERS:
            content = sections[marker]['content']
            patch_name = sections[marker]['patch']

            final_script += f'## 📝 Applied from patch "{patch_name}"'
            final_script += '\n'
            final_script += content
            final_script += '\n'

        final_script += '\n\napp.publish()\n```'
        return final_script

    def _first_iteration(self):
        return len(self._conversation) == 0

    def _build_system_prompt(self, create_or_edit):

        print('Using the prompt to ' + create_or_edit)
        main_task = self._load_prompt_file(create_or_edit, 'main-task')
        kawa_sdk_documentation = self._load_prompt_file('common', 'sdk-documentation')
        mistakes_to_avoid = self._load_prompt_file('common', 'mistakes-to-avoid')
        example = self._load_prompt_file(create_or_edit, 'examples')
        guidelines = self._load_prompt_file('common', 'additional-guidelines')

        system_prompt = f'''

        # Main task

        {main_task}


        # Kawa SDK Documentation

        {kawa_sdk_documentation}


        # Additional guidelines
        
        {guidelines}
        
        
        # Mistakes to avoid

        {mistakes_to_avoid}


        # Full working example

        {example}
        '''

        return system_prompt

    def _generate_content_open_ai(self, client, model, system_prompt):
        conversation_with_system_prompt = [{"role": "system", "content": system_prompt}] + self._conversation
        generated_chunks = []
        # Stream tokens from Qwen3 (OpenAI-compatible Chat Completions API)
        stream = client.chat.completions.create(
            model=model,
            messages=conversation_with_system_prompt,
            stream=True,
        )

        for chunk in stream:
            # OpenAI-compatible delta structure
            delta = None
            if chunk.choices and chunk.choices[0].delta:
                delta = chunk.choices[0].delta
            if delta and getattr(delta, "content", None):
                text = delta.content
                generated_chunks.append(text)
                print(text, end="", flush=True)

        return "".join(generated_chunks)

    def _generate_content_claude(self, system_prompt):
        generated_content = ''
        with self._anthropic_client.messages.stream(
                model=MODEL,
                max_tokens=50_000,
                system=system_prompt,
                messages=self._conversation,
        ) as stream:
            for text in stream.text_stream:
                generated_content += text
                print(text, end="", flush=True)

        return generated_content

    @staticmethod
    def _load_prompt_file(directory, name):
        full_path = f'{script_dir}/prompts/{directory}/{name}.md'
        print(f'Loading prompt:{name} from {full_path}')

        with open(full_path, "r", encoding="utf-8") as file:
            prompt = file.read()

        if not prompt:
            raise Exception('Could not load ' + full_path)

        return prompt


class StreamDuplicator:
    def __init__(self, *streams):
        self.streams = streams

    def write(self, data):
        for stream in self.streams:
            stream.write(data)
            stream.flush()

    def flush(self):
        for stream in self.streams:
            stream.flush()



import datetime


def date_to_epoch_day(date: datetime.date) -> int:
    return (date - datetime.date(1970, 1, 1)).days if date else None


def datetime_to_time_stamp_millis(date_time: datetime.datetime) -> int:
    return int(date_time.timestamp() * 1000) if date_time else None



import os
import shutil
import glob
import json
import io
import re
import importlib
import uuid
import sys
import time
from openai import OpenAI
from anthropic import Anthropic
from datetime import datetime
from kywy.client.application.chat.source_code_utils import extract_before_marker_line, extract_between_marker_lines, \
    extract_sections, extract_between_fence

open_ai_client = OpenAI()
anthropic_client = Anthropic()
ROOT_DIR = os.environ['CO_BUILDER_ROOT_DIR']
SCRIPT_DIR = ROOT_DIR + '/prompts'
CONTEXT_ROOT = ROOT_DIR + '/data'
CODER_AGENT = 'coder_agent'
SPECS_AGENT = 'specs_agent'
MARKERS = [
    '-- DATA SECTION',
    '-- RELATIONSHIPS SECTION',
    '-- VARIABLES SECTION',
    '-- METRICS SECTION',
    '-- DASHBOARD SECTION'
]


def init(application_id, reset=False):
    path = CONTEXT_ROOT + '/' + application_id

    if reset and os.path.exists(path) and os.path.isdir(path):
        shutil.rmtree(path)

    os.makedirs(_agent_dir(application_id, CODER_AGENT), exist_ok=False)
    os.makedirs(_agent_dir(application_id, SPECS_AGENT), exist_ok=False)


def publish(application_id, step_name=None, num_retries=2, print_output=False, running_on_dev_server=False):
    for try_id in range(0, num_retries):
        print(f'🐢 Now running the publication script (Iteration={try_id + 1}/{num_retries + 1})')
        feedback = _publish(
            application_id=application_id,
            step_name=step_name,
            print_output=print_output,
            running_on_dev_server=running_on_dev_server,
        )
        if feedback:
            print('Generating corrective patch')
            generate_code_patch(
                application_id=application_id,
                user_prompt=feedback,
                step_instructions='Focus on fixing the reported errors',
            )
        else:
            return


def generate_datasets(application_id, special_instructions=""):
    generate_code_patch(
        application_id=application_id,
        user_prompt=special_instructions,
        step_instructions="""
        Only generate the datasets
        """,
    );


def generate_relationships(application_id, special_instructions=""):
    generate_code_patch(
        application_id=application_id,
        user_prompt=special_instructions,
        step_instructions="""
        Add the relationships in the RELATIONSHIPS SECTION.
        """,
    );


def generate_variables(application_id, special_instructions=""):
    generate_code_patch(
        application_id=application_id,
        user_prompt=special_instructions,
        step_instructions="""
        Add the variables in the VARIABLES SECTION.
        """,
    );


def generate_metrics(application_id, special_instructions=""):
    generate_code_patch(
        application_id=application_id,
        user_prompt=special_instructions,
        step_instructions="""
        Add the metrics in the METRICS SECTION.
        """,
    );


def generate_dshboards(application_id, special_instructions=""):
    generate_code_patch(
        application_id=application_id,
        user_prompt=special_instructions,
        step_instructions="""
        Add the widget and pages in the DASHBOARD SECTION.
        """,
    );


def generate_specs(application_id,
                   user_prompt=None,
                   predefined_specs=None,
                   print_output=False):

    if predefined_specs:
        predefined_specs = predefined_specs.strip()

    if user_prompt:
        user_prompt = user_prompt.strip()

    if not user_prompt and not predefined_specs:
        raise Exception('We need at least one of user_prompt or predefined_specs')

    if predefined_specs:
        specs = predefined_specs
    else:
        conversation = _restore_conversation(
            application_id=application_id,
            agent_id=SPECS_AGENT,
        )

        system_prompt = _read_prompt('research', 'research')
        specs = _generate_content_open_ai(
            conversation=conversation,
            system_prompt=system_prompt,
            user_prompt=user_prompt,
            print_output=print_output,
        )

        _save_conversation(
            application_id=application_id,
            agent_id=SPECS_AGENT,
            conversation=conversation,
        )

    timestamp = datetime.now().isoformat(timespec="seconds")
    for filename in ['specs.md', f'specs-{timestamp}.md']:
        _persist_output_as_file(
            application_id=application_id,
            agent_id=SPECS_AGENT,
            filename=filename,
            content=specs,
        )


def generate_code_patch(application_id,
                        user_prompt,
                        step_instructions,
                        print_output=False):
    conversation = _restore_conversation(
        application_id=application_id,
        agent_id=CODER_AGENT,
    )

    action = 'create' if not conversation else 'edit'

    system_prompt = _code_generation_prompt(action)

    specs = _load_file(
        application_id=application_id,
        agent_id=SPECS_AGENT,
        filename='specs.md'
    )

    patch = _generate_content_anthropic(
        print_output=print_output,
        conversation=conversation,
        system_prompt=system_prompt,
        user_prompt=f"""

        {step_instructions}

        User prompt: {user_prompt}

        Specs: {specs}

        """,
    )

    patch = extract_between_fence(patch)

    _save_conversation(
        application_id=application_id,
        agent_id=CODER_AGENT,
        conversation=conversation,
    )

    timestamp = datetime.now().isoformat(timespec="seconds")

    filename = 'base.py' if action == 'create' else f'patch-{timestamp}.py'

    _persist_output_as_file(
        application_id=application_id,
        agent_id=CODER_AGENT,
        filename=filename,
        content=patch,
    )

    final_script = _merge_patches(
        application_id=application_id
    )

    _persist_output_as_file(
        application_id=application_id,
        agent_id=CODER_AGENT,
        filename='merged.py',
        content=final_script,
    )

    return patch


def _code_generation_prompt(create_or_edit):
    main_task = _read_prompt(create_or_edit, 'main-task')
    kawa_sdk_documentation = _read_prompt('common', 'sdk-documentation')
    mistakes_to_avoid = _read_prompt('common', 'mistakes-to-avoid')
    example = _read_prompt(create_or_edit, 'examples')
    guidelines = _read_prompt('common', 'additional-guidelines')

    system_prompt = f'''

        # Main task

        {main_task}


        # Kawa SDK Documentation

        {kawa_sdk_documentation}


        # Additional guidelines

        {guidelines}


        # Mistakes to avoid

        {mistakes_to_avoid}


        # Full working example

        {example}
        '''

    return system_prompt


def _read_prompt(directory, name):
    full_path = f'{SCRIPT_DIR}/{directory}/{name}.md'
    with open(full_path, "r", encoding="utf-8") as file:
        prompt = file.read()

    if not prompt:
        raise Exception('Could not load ' + full_path)

    return prompt


def _merge_patches(application_id):
    base_script = _load_file(application_id, CODER_AGENT, 'base.py')

    # Init the sections with the base file
    sections = extract_sections(source=base_script, markers=MARKERS)
    for marker in MARKERS:
        if marker not in sections:
            raise Exception(f'The initial script is missing this section: {marker}')

    # Go through all the files in order and apply the patches, updating the "sections" dict
    patch_filenames = _list_files_in_dir(
        application_id=application_id,
        agent_id=CODER_AGENT,
        pattern='patch-*.py',
    )

    for patch_filename in patch_filenames:
        patch_content = _load_file(
            application_id=application_id,
            agent_id=CODER_AGENT,
            filename=patch_filename,
        )

        sections_in_patch = extract_sections(
            source=patch_content,
            markers=MARKERS,
            patch_name=patch_filename,
        )

        sections.update(sections_in_patch)

    return _build_script_from_sections(
        base_script=base_script,
        sections=sections,
    )


def _build_script_from_sections(base_script, sections):
    script_header = extract_before_marker_line(
        source=base_script,
        marker=MARKERS[0]
    )
    final_script = script_header

    for marker in MARKERS:
        content = sections[marker]['content']
        patch_name = sections[marker]['patch']
        final_script += f'## 📝 Applied from patch "{patch_name}"'
        final_script += '\n'
        final_script += content
        final_script += '\n'

    final_script += '\n\napp.publish()\n'
    return final_script


def _publish(application_id, step_name=None, print_output=False, running_on_dev_server=True):
    agant_dir = _agent_dir(application_id=application_id, agent_id=CODER_AGENT, )
    script_path = f'{agant_dir}/merged.py'
    stdout_buffer = io.StringIO()
    original_stdout = sys.stdout

    if print_output:
        sys.stdout = StreamDuplicator(original_stdout, stdout_buffer)
    else:
        sys.stdout = stdout_buffer

    error = None
    try:
        module_name = f"mod_{uuid.uuid4().hex}"
        spec = importlib.util.spec_from_file_location(module_name, script_path)
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)

    except Exception as e:
        error = f'An error was detected: {e}'
        print(error)

    finally:
        sys.stdout = original_stdout
        captured_output = stdout_buffer.getvalue()
        error_section = extract_between_marker_lines(
            captured_output,
            marker='# ERROR SECTION'
        ).strip()

        if error:
            return f"""
            There was a global error, please regenerate the appropriate sections to fix it.
            {error}
            """

        elif error_section:
            return f"""
            Here is the output of the script - notice that it stopped in the middle of its execution.
            This means that the last element has an error to fix (indicated below).
            Please generate ONLY the impacted section with your fix.
            {captured_output}
            """
        else:
            for line in captured_output.splitlines():
                if 'Publication Complete:' in line:
                    if running_on_dev_server:
                        updated_string = line.replace("8080", "4200")
                        print(updated_string)
                    else:
                        print(line)

                # If a step name was specified, we will extract the report for that step
                # and print it out
                elif step_name and 'Report file here' in line:
                    path = re.findall(r'\[(.*?)\]', line)[0]
                    with open(path, "r") as report_file:
                        report = json.loads(report_file.read())
                    if report:
                        created_entities = report.get('report', {}).get(step_name, [])
                        print(f'\nThe following {step_name}s were created:')
                        for created_entity in created_entities:
                            print(' - ' + created_entity.get('name'))

            return ''


## 🤖 LLM calls

def _generate_content_open_ai(conversation,
                              system_prompt,
                              user_prompt,
                              model=None,
                              print_output=True):
    start = time.time()
    conversation.append({"role": "user", "content": user_prompt})
    system_msg = {"role": "system", "content": system_prompt}
    full_conversation = [system_msg] + conversation
    used_model =model or 'gpt-4o'
    stream = open_ai_client.chat.completions.create(
        model=used_model,
        messages=full_conversation,
        stream=True
    )

    agent_message = ''
    counter = 0
    for event in stream:
        if event.choices[0].delta.content is not None:
            delta = event.choices[0].delta.content
            agent_message += delta
            if print_output:
                print(delta, end="")
            else:
                counter += 1
                print(f"\rReceiving content: {counter}", end="")

    conversation.append({"role": "assistant", "content": agent_message})
    elapsed = time.time() - start
    print(f'\n✅ ({used_model}) Generation is complete (it contains: {len(agent_message)} characters in {elapsed:.2f}s)')

    return agent_message


def _generate_content_anthropic(conversation,
                                system_prompt,
                                user_prompt,
                                model=None,
                                print_output=True):
    start = time.time()
    conversation.append({"role": "user", "content": user_prompt})
    agent_message = ''
    counter = 0
    used_model = model or 'claude-sonnet-4-20250514'
    with anthropic_client.messages.stream(
            model=used_model,
            max_tokens=50_000,
            system=system_prompt,
            messages=conversation,
    ) as stream:
        for text in stream.text_stream:
            agent_message += text
            if print_output:
                print(text, end="", flush=True)
            else:
                counter += 1
                print(f"\rReceiving content: {counter}", end="")

    conversation.append({"role": "assistant", "content": agent_message})
    elapsed = time.time() - start
    print(f'\n✅ ({used_model}) Generation is complete (it contains: {len(agent_message)} characters in {elapsed:.2f}s)')

    return agent_message


## 📝 Simple file manipulation

def _agent_dir(application_id, agent_id):
    return CONTEXT_ROOT + '/' + application_id + '/' + agent_id


def _list_files_in_dir(application_id, agent_id, pattern):
    directory = _agent_dir(application_id, agent_id)
    path_pattern = os.path.join(directory, pattern)
    file_paths = glob.glob(path_pattern)
    filenames = [os.path.basename(path) for path in file_paths]
    return sorted(filenames)


def _persist_output_as_file(application_id, agent_id, filename, content):
    path = _agent_dir(application_id, agent_id) + '/' + filename
    with open(path, "w") as f:
        f.write(content)


def _load_file(application_id, agent_id, filename):
    path = _agent_dir(application_id, agent_id) + '/' + filename
    try:
        with open(path, "r") as f:
            return f.read()
    except FileNotFoundError:
        return ""


def _save_conversation(application_id, agent_id, conversation):
    path = _agent_dir(application_id, agent_id) + '/' + "messages.json"
    with open(path, "w") as f:
        json.dump(conversation, f)


def _restore_conversation(application_id, agent_id):
    path = _agent_dir(application_id, agent_id) + '/' + "messages.json"
    try:
        with open(path, "r") as f:
            return json.load(f)
    except (FileNotFoundError, json.JSONDecodeError):
        return []


class StreamDuplicator:
    def __init__(self, *streams):
        self.streams = streams

    def write(self, data):
        for stream in self.streams:
            stream.write(data)
            stream.flush()

    def flush(self):
        for stream in self.streams:
            stream.flush()



import datetime

from .kawa_client import KawaClient

from .generator import Generators as g
import pandas as pd
from pathlib import Path
import pyarrow as pa
import pyarrow.parquet as pq


class DataGenerator:

    def __init__(self,
                 kawa_client: KawaClient,
                 sheet_and_datasource_name,
                 number_of_rows=10,
                 number_of_dimensions=5,
                 number_of_measures=5):
        """
        Initializes the data generator

        :param kawa_client: the kawa client
        :param sheet_and_datasource_name: the name of the datasource and sheet to create
        :param number_of_rows: how many rows to feed
        :param number_of_dimensions: how many extra dimensions
        :param number_of_measures: how many extra measure columns
        """
        self._k = kawa_client
        self._sheet_and_datasource_name = sheet_and_datasource_name
        self._number_of_rows = number_of_rows
        self._number_of_dimensions = number_of_dimensions
        self._number_of_measures = number_of_measures

    def generate_data(self):
        df = self._generate_df()
        data_loader = self._k.new_data_loader(df, self._sheet_and_datasource_name)
        data_loader.create_datasource(primary_keys=['f_id'])

        nb_files = max(1, int(self._number_of_rows / (100 * 1000)))
        tmp_dir = self._k.tmp_files_directory
        print("Will generate {} parquet files".format(nb_files))
        try:
            all_files = []
            for i in range(nb_files):
                filename = '{}/kawa.{}.parquet'.format(tmp_dir, i + 1)
                print('Exporting file ' + filename)
                df['f_id'] = df.index + i * self._number_of_rows
                df['f_pseudo_id_2'] = df['f_id'].map(lambda x: 'pseudo-k2-' + str(int(x / 2)))
                df['f_pseudo_id_3'] = df['f_id'].map(lambda x: 'pseudo-k3-' + str(int(x / 3)))
                df['f_pseudo_id_4'] = df['f_id'].map(lambda x: 'pseudo-k4-' + str(int(x / 4)))
                df['f_pseudo_id_5'] = df['f_id'].map(lambda x: 'pseudo-k5-' + str(int(x / 5)))
                df['f_pseudo_id_6'] = df['f_id'].map(lambda x: 'pseudo-k10-' + str(int(x / 10)))
                df['f_id'] = df['f_id'].apply(str)
                pq.write_table(pa.Table.from_pandas(df=df), filename)
                all_files.append(filename)

            nb_threads = min(20, len(all_files))
            data_loader.load_data(create_sheet=True,
                                  reset_before_insert=True,
                                  nb_threads=nb_threads,
                                  parquet_file_list=all_files)
        finally:
            for p in Path(tmp_dir).glob("kawa.*.parquet"):
                p.unlink()

    def _generate_df(self):
        data = []
        print("Generating the data")

        # The df contains at most 100,000 rows
        nb_rows_in_df = min(self._number_of_rows, 100 * 1000)

        for i in range(nb_rows_in_df):
            row = self._one_row_of_data(i)
            data.append(row)

        df = pd.DataFrame.from_records(data)

        return df

    def _one_row_of_data(self, row_id):
        number_of_dimension_columns = self._number_of_dimensions
        number_of_measure_columns = self._number_of_measures
        random_date = datetime.date(2024, 1, g.generate_random_int(1, 10))

        base = {
            # id and pseudo ids will be overridden further on
            'f_id': str(row_id),
            'f_pseudo_id_2': 'a2',
            'f_pseudo_id_3': 'a3',
            'f_pseudo_id_4': 'a4',
            'f_pseudo_id_5': 'a5',
            'f_pseudo_id_6': 'a6',

            'f_date': random_date,
            'f_text': 'text-{}'.format(g.generate_random_int(1, 10000)),
            'f_dimension_10': 'dim-{}'.format(g.generate_random_int(1, 10)),
            'f_dimension_100': 'dim-{}'.format(g.generate_random_int(1, 100)),
            'f_dimension_1000': 'dim-{}'.format(g.generate_random_int(1, 1000)),

            'system': 'text-{}'.format(g.generate_random_int(1, 5)),
            'isin': 'isin-{}'.format(row_id),
            'trading_region': 'region{}'.format(g.generate_random_int(1, 3)),
            'strategy': 'strat{}'.format(g.generate_random_int(1, 200)),
            'book': 'book{}'.format(g.generate_random_int(1, 1000)),
            'portfolio': 'ptf{}'.format(g.generate_random_int(1, 3000)),


        }

        base['netting-key-7'] = '{}+{}+{}+{}+{}+{}+{}'.format(
            base['f_date'],
            base['system'],
            base['isin'],
            base['trading_region'],
            base['strategy'],
            base['book'],
            base['portfolio'],
        )

        base['netting-key-6'] = '{}+{}+{}+{}+{}+{}'.format(
            base['f_date'],
            base['system'],
            base['isin'],
            base['trading_region'],
            base['strategy'],
            base['book']
        )

        base['netting-key-5'] = '{}+{}+{}+{}+{}'.format(
            base['f_date'],
            base['system'],
            base['isin'],
            base['trading_region'],
            base['strategy']
        )

        base['netting-key-4'] = '{}+{}+{}+{}'.format(
            base['f_date'],
            base['system'],
            base['isin'],
            base['trading_region']
        )

        for i in range(number_of_dimension_columns):
            base['dim_{}'.format(i)] = 'a-{}'.format(g.generate_random_int(1, 1000))

        for i in range(number_of_measure_columns):
            base['meas_{}'.format(i)] = g.generate_random_int(1, 1000)

        return base



class Types:
    TEXT = 'text'
    DECIMAL = 'decimal'
    DATE = 'date'
    DATE_TIME = 'date_time'
    BOOLEAN = 'boolean'

    LIST_OF_DECIMALS = 'list(integer,decimal)'
    LIST_OF_TEXTS = 'list(integer,text)'

    TIME_SERIES_OF_DECIMALS = 'timeseries(date,decimal)'



import os
import sys
from . import __version__

from kywy.client.kawa_client import KawaClient


def main():
    url = os.getenv("KAWA_URL")
    username = os.getenv("KAWA_USER")
    password = os.getenv("KAWA_PASSWORD")

    print(f"Kywy version: {__version__}")

    if not url or not username or not password:
        print("Variables are missing, silently exiting")
        sys.exit(0)

    kawa_root = KawaClient(url)
    kawa_root.login_with_credential(username, password)

    root_user = kawa_root.get_current_user()
    root_user_email = root_user.get("email")
    print(f'Successfully logged in as {root_user_email}')


if __name__ == "__main__":
    main()



import pandas as pd

from ...client.kawa_client import KawaClient as K
from datetime import datetime, date, timedelta
import numpy as np
from faker import Faker
import json


def kawa():
    k = K(kawa_api_url='http://localhost:8080')
    k.login_with_credential(login='setup-admin@kawa.io', password='changeme')
    k.set_active_workspace_id(workspace_id='107')
    return k


research = kawa().research('Some research')
orders_model = research.register_model('4727')
events_model = research.register_model('4728')
main_model = orders_model

events_relationship = orders_model.create_relationship(
    name='Event per State..',
    description="""
      We are joining the events to the orders to get the cost per state on the State.
      """,
    origin_model=events_model,
    link={'State': 'event state'},
)

events_relationship.add_column(
    origin_column='event cost',
    aggregation='SUM',
    new_column_name='Cost per State..',
)

orders_relationship = orders_model.create_relationship(
    name='II Cost per State',
    description="""
      We are joining the orders to the event to get the profit per state on the State
      """,
    target_model=orders_model,
    link={'State': 'State'},
)

orders_relationship.add_column(
    name='Profit',
    aggregation='SUM',
    new_column_name='Profit per State..',
    filters=[
        K.col('State').in_list('Ohio', 'Maine')
    ]
)

orders_model.create_metric(
    name='II This is FOO',
    formula='"Profit" / "Quantity"'
)

orders_model.create_metric(
    name='II This is BAR',
    formula='"II This is FOO" / 2 '
)

print(research.publish_main_model(main_model=main_model))

# research.bar_chart(
#     title='Profit per State',
#     x='State',
#     y='Unit Profit',
#     color='Segment',
#     show_values=True,
#     show_totals=True,
#     model=orders_model,
#     filters=[
#         # K.col("State").in_list("California", "Ohio"),
#         K.col("Unit Profit").gt(1)
#     ],
#     order_by='Unit Profit',
#     order_direction='DESCENDING',
#     limit=5,
# )
#
# research.scatter_chart(
#     title='Order Count vs. Profit by Month',
#     granularity='Order Date',
#     x='Order ID',
#     aggregation_x='COUNT',
#     y='Profit',
#     aggregation_y='SUM',
#     color='Profit',
#     aggregation_color='SUM',
#     model=orders_model,
#     time_sampling='YEAR_AND_MONTH',
# )
#
# df = (orders_model
#       .select(
#     K.col('State'),
#     K.col('Unit Profit III').avg().alias('Average profit'),
#     K.col('Cost per State').sum().alias('Total per state'),
#     K.col('Profit per State').median().alias('Median profit per state'),
# )
#       .group_by('State')
#       .query_description('This is the description')
#       .collect())
#
# research.register_result(
#     description='Dataframe containing data about cost and profit',
#     df=df
# )
# print(research.publish_results())



import time
import traceback

from typing import Callable

from .app_reporter import Reporter
from .app_agent import Agent
from .app_dataset import DataSet
from .app_filter import TextFilter
from .app_model import DataModel
from .app_page import ApplicationPage
from .application_building_utils import info, start_sync, \
    error, feedback, PALETTES, init_logs


class ApplicationBuilder:

    def __init__(self,
                 kawa_client,
                 name,
                 sidebar_color=None,
                 palette=None,
                 palette_name=None,
                 create_default_components=True,
                 unique_tag=None):
        self._k = kawa_client
        self._reporter = Reporter(name=name)
        self._name = name
        self._datasets = []
        self._pages = []
        self._text_filters = {}
        self._model = None
        self._application_id = None
        self._application_url = None
        self._agent = None
        self._unique_tag = unique_tag or f'#{name}'
        self._sidebar_color = sidebar_color
        self._create_default_components = create_default_components

        init_logs()

        if palette_name:
            self._palette = PALETTES.get(palette_name)
        else:
            self._palette = palette

    @property
    def application_id(self):
        return self._application_id

    @property
    def tag(self):
        return self._unique_tag

    @property
    def agent_id(self):
        if self._agent:
            return self._agent.agent_id

    @property
    def data_is_configured(self):
        return bool(self._model)

    @property
    def url(self):
        return self._application_url

    def create_model(self, dataset):
        self._model = DataModel(
            kawa=self._k,
            reporter=self._reporter,
            name=self._name,
            dataset=dataset,
        )
        return self._model

    def create_dataset(self, name: str, generator: Callable = None, datasource_id=None):
        existing_names = [d.name for d in self._datasets]
        if name in existing_names:
            raise Exception(f'A dataset with the name {name} already exists. Please pick a different one')

        dataset = DataSet(
            kawa=self._k,
            reporter=self._reporter,
            application=self,
            name=name,
            generator=generator,
        )
        self._datasets.append(dataset)
        return dataset

    def create_page(self, name: str):
        existing_names = [d.name for d in self._pages]
        if name in existing_names:
            raise Exception(f'A page with the name {name} already exists. Please pick a different one')

        page = ApplicationPage(
            kawa=self._k,
            reporter=self._reporter,
            data_model=self._model,
            application=self,
            name=name,
        )
        self._pages.append(page)
        return page

    def create_text_filter(self, name, filtered_column=None, source=None):
        # Merges filters with the same name
        if name not in self._text_filters:
            self._text_filters[name] = TextFilter(
                kawa=self._k,
                reporter=self._reporter,
                name=name,
            )

        self._text_filters[name].append_column(
            sheet_id_supplier=lambda: source.sheet_id if source else self._model.sheet_id,
            # If filtered column is not specified, rever to name
            column_name=filtered_column or name,
        )

    def create_ai_agent(self, name, instructions, color=None):
        self._agent = Agent(
            kawa=self._k,
            reporter=self._reporter,
            name=name,
            instructions=instructions,
            application=self,
            color=color,
        )

    def publish(self):

        info('---' * 30)
        info(f'🚀 Publishing app "{self._name}" to {self._k.kawa_api_url}')
        info('---' * 30)
        export_file = self._reporter.export()
        try:
            self.sync()
            self._health_check()
        except Exception as e:
            error_message = str(e)
            trace = traceback.format_exc()
            error(f'{error_message} - {trace}')
        finally:
            if self.data_is_configured:
                export_file = self._reporter.export()
                if export_file:
                    feedback(self.url, report_file=export_file)

    def sync(self):
        start_sync(f'Application {self._name}')

        existing_application = self._load_state()
        if existing_application:
            self._cleanup_existing_application(existing_application)

        if self._agent:
            self._agent.sync()

        if not self._model and not self._datasets:
            info('👻 No data configured')
            return

        for dataset in self._datasets:
            dataset.sync()

        if not self._model:
            # By default, init the model on the first dataset
            self.create_model(self._datasets[0])

        self._model.sync()

        if not self._model.sheet_id:
            raise Exception('The underlying model has not been synced')

        if existing_application:
            self._update_application_if_necessary(existing_application)
        else:
            self._create_new_application()

        extended_application = self._k.entities.extended_applications().get_entity_by_id(self._application_id)
        for text_filter in self._text_filters.values():
            text_filter.sync(
                extended_application=extended_application,
            )

        for page in self._pages:
            page.sync(application_id=self._application_id)

    def _load_state(self):
        return self._k.entities.applications().find_entity_by_tag(tag=self.tag)

    def _health_check(self, iteration=1):

        if not self.data_is_configured:
            return

        max_iter = 10
        if iteration > max_iter:
            error(f'🚨 Application is not healthy after allocated wait time')

        num_datasets = len(self._datasets)
        failures = []

        if iteration == 1:
            ds = 'datasets' if num_datasets > 1 else 'dataset'
            info(f'🚑 Running healthcheck on {num_datasets} {ds}')
            for dataset in self._datasets:
                info(f'--> {dataset.name}')

            time.sleep(1)
        else:
            info(f'🚑 Retry {iteration - 1}/{max_iter}')
            time.sleep(iteration)

        for dataset in self._datasets:
            ds_id = str(dataset.datasource_id)
            url = f'{self._k.kawa_api_url}/backoffice/datasources/health-report/v2/{ds_id}'
            health_reports = self._k.get(url)
            chronological = sorted(health_reports, key=lambda x: x['synchronizationState']['startTime'])
            last_report = chronological[-1]
            status = last_report['synchronizationState']['status']

            if status == 'SUCCESS':
                info(f'💚 Dataset {dataset.name} is healthy')
            elif status == 'RUNNING':
                info(f'🕣 Dataset {dataset.name} is still running - let\'s wait a bit and check again.')
                self._health_check(iteration + 1)
            elif status == 'FAILURE':
                error(f'🚨 Dataset {dataset.name} is NOT healthy')
                error(last_report['synchronizationState']['logAnswer']['log'])
                failures.append(dataset.name)

        if failures:
            error(f'🚨 Application is not healthy')

    def _update_application_if_necessary(self, existing_application):
        workspace_id = existing_application['workspaceId']
        base_url = self._k.kawa_api_url
        self._application_id = existing_application['id']
        self._application_url = f'{base_url}/workspaces/{workspace_id}/applications/{self._application_id}'

        required_sheet_ids = [d.sheet_id for d in self._datasets]
        self._k.commands.run_command('replaceApplicationSheetIds', {
            "applicationId": str(self._application_id),
            "sheetIds": list(required_sheet_ids),
            "createDefaultComponents": True,
            "forceCreateAllDefaultComponents": True,
        })

    def _create_new_application(self):
        if self._palette and len(self._palette) > 7:
            self._k.commands.run_command('replaceWorkspacePalette', {
                "workspaceId": self._k.active_workspace_id,
                "palette": {
                    "colors": self._palette,
                    "enabled": True
                }
            })

        created_app = self._k.commands.run_command('createApplication', {
            "displayInformation": {
                "displayName": self._name,
                "description": "",
                "extraInformation": {
                    "immutableTag": f'{self.tag}'
                }
            },
            "sheetIds": [d.sheet_id for d in self._datasets],
            "agentIds": [self.agent_id] if self.agent_id else [],
            "createDefaultComponents": self._create_default_components,
        })['application']

        workspace_id = created_app['workspaceId']
        base_url = self._k.kawa_api_url
        self._application_id = created_app['id']
        self._application_url = f'{base_url}/workspaces/{workspace_id}/applications/{self._application_id}'
        info(f'⚙️ Application {self._name} was created (id={self._application_id})')
        info(f'🏷️ New tag for application: {self._unique_tag}')

        if self._sidebar_color:
            self._k.commands.run_command('replaceApplicationDisplayParameters', {
                "applicationId": str(self._application_id),
                "displayParameters": {
                    "color": self._sidebar_color,
                }
            })

        return self._application_id

    def _cleanup_existing_application(self, existing_application):
        view_component_list = [
            str(c['componentId'])
            for c in existing_application.get('components', [])
            if c['componentType'] == 'VIEW'
        ]
        for component_id in view_component_list:
            print('Removing component')
            self._k.commands.run_command('removeComponentFromApplication', {
                "applicationId": str(existing_application['id']),
                "componentId": component_id,
            })



from .application_building_utils import get_column, load_sheet
from .app_synchronizer import Synchronizer
import re

ACCEPTABLE_AGGREGATIONS = [
    'FIRST',
    'SUM',
    'AVERAGE',
    'COUNT',
    'MEDIAN',
    'COUNT_UNIQUE',
    'MIN',
    'MAX'
]


class Relationship:

    def __init__(self, kawa, reporter, name, model, link,
                 description=None, dataset=None, target_sheet=None, ai_mode=False):
        self._k = kawa
        self._reporter = reporter
        self._name = name
        self._description = description
        self._dataset = dataset
        self._target_sheet = target_sheet
        self._model = model
        self._link = link
        self._cached_sheet = None
        self._columns = []
        self._ai_mode = ai_mode

    @property
    def source_sheet(self):
        return self._model.sheet

    @property
    def target_sheet(self):
        return self._target_sheet if self._target_sheet else self._dataset.sheet

    @property
    def name(self):
        return self._name

    def add_column(self,
                   aggregation,
                   new_column_name,
                   filters=None,
                   name=None,
                   origin_column=None):
        uc_aggregation = aggregation.upper()
        uc_aggregation = 'FIRST' if uc_aggregation == 'ANY_VALUE' else uc_aggregation
        if uc_aggregation not in ACCEPTABLE_AGGREGATIONS:
            raise Exception('The aggregation is not known, please use one of: ' + ','.join(ACCEPTABLE_AGGREGATIONS))

        self._columns.append({
            'name': name or origin_column,
            'aggregation': uc_aggregation,
            'new_column_name': new_column_name,
            'filters': filters,
        })

    def sync(self):
        if not self._columns:
            return

        if not self._link:
            raise Exception('There is no definition for the link in this relationship')

        for column in self._columns:
            Relationship._Synchronizer(
                kawa=self._k,
                ai_mode=self._ai_mode,
                relationship=self,
                column=column
            ).sync()

        self._reporter.report(
            object_type='Relationship',
            name=self._name,
        )

    def build_join_definition(self):
        joins = []
        for source, target in self._link.items():
            source_column = get_column(
                sheet=self.source_sheet,
                column_name=source,
                kawa=self._k,
                force_refresh_sheet=True
            )

            target_column = get_column(
                sheet=self.target_sheet,
                column_name=target,
                kawa=self._k,
                force_refresh_sheet=True
            )

            joins.append({
                "targetColumnId": target_column['columnId'],
                "sourceColumnId": source_column['columnId'],
            })
        return joins

    def to_ascii(self):

        source_sheet_name = remove_uuid_suffix(self.source_sheet['displayInformation']['displayName'])
        target_sheet_name = remove_uuid_suffix(self.target_sheet['displayInformation']['displayName'])

        on_clause_collector = []

        for source, target in self._link.items():
            on_clause = f'"{source_sheet_name}"."{source}"="{target_sheet_name}"."{target}"'
            on_clause_collector.append(on_clause)

        on_clauses = ' and '.join(on_clause_collector)
        ascii_representation = f'"{source_sheet_name}" LINKED WITH "{target_sheet_name}" ON ({on_clauses}):\n'

        ascii_representation += "  |\n"
        for col in self._columns:
            col_ascii = f"  |-> {col['aggregation']}({col['name']})\n"
            ascii_representation += col_ascii

        return ascii_representation.strip()

    class _Synchronizer(Synchronizer):
        def __init__(self, kawa, ai_mode, relationship, column):
            super().__init__(
                kawa=kawa,
                icon='🔗',
                entity_description=f'Relationship "{relationship.name}"',
            )
            self._relationship = relationship
            self._column = column
            self._ai_mode = ai_mode

        def _load_state(self):
            if self._ai_mode:
                # State is not required in AI Mode, we will always be adding new adhoc columns
                return {}
            else:
                existing_columns = {
                    c['displayInformation']['displayName']: c
                    for c in self._relationship.source_sheet['computedColumns']
                    if c['columnNature'] == 'LINKED' and c['columnStatus'] == 'ACTIVE'
                }
                return existing_columns.get(self._column['new_column_name'])

        def _raise_if_state_invalid(self):
            ...

        def _should_create(self):
            if self._ai_mode:
                # Always create in AI Mode, because we create adHoc columns
                return True
            else:
                return self._state is None

        def _create_new_entity(self):
            source_sheet = self._relationship.source_sheet
            target_sheet = self._relationship.target_sheet

            target_column = get_column(
                sheet=target_sheet,
                column_name=self._column['name'],
                kawa=self._k,
                force_refresh_sheet=True
            )

            joins = self._relationship.build_join_definition()

            filters = [f.to_dict() for f in (self._column.get('filters') or [])]
            self._k.commands.run_command(
                command_name='addLookupField',
                command_parameters={
                    "filters": filters,
                    "layoutId": str(source_sheet['defaultLayoutId']),  # This is the source layout
                    "linkedSheetId": str(target_sheet['id']),  # This is the target sheet
                    "columnDefinitions": [
                        {
                            "columnId": target_column['columnId'],
                            "aggregation": self._column['aggregation'],
                            "lookupColumnName": self._column['new_column_name'],
                        }
                    ],
                    "joins": joins,
                    "adHoc": self._ai_mode,
                }
            )

        def _update_entity(self):
            existing_lookup_column = self._state
            existing_joins = existing_lookup_column['joins']
            new_joins = self._relationship.build_join_definition()

            existing_joins_for_comparison = sorted([(j['sourceColumnId'], j['targetColumnId']) for j in existing_joins])
            new_joins_for_comparison = sorted([(j['sourceColumnId'], j['targetColumnId']) for j in new_joins])
            if existing_joins_for_comparison != new_joins_for_comparison:
                # TODO We need to update the join for this column
                ...

            # TODO: Update aggregation -> Need to upgrade definition view

            # TODO: Update the target column (We need BE support for this)

            ...

        def _build_new_state(self):
            pass


def remove_uuid_suffix(s: str) -> str:
    return re.sub(r'\s*\([0-9a-fA-F\-]{36}\)\s*$', '', s)



import uuid
from dataclasses import dataclass, asdict


@dataclass
class ReportBlock:
    id: str
    type: str
    data: dict[str, any]

    @staticmethod
    def content(content_type, content_name, content):
        return ReportBlock(
            id=ReportBlock.generate_random_id(),
            type=content_type,
            data={
                content_name: content
            }
        )

    @staticmethod
    def header(level, content):

        if level == 1:
            header_type = 'headerOne'
        elif level == 2:
            header_type = 'headerTwo'
        elif level == 3:
            header_type = 'headerThree'
        else:
            raise Exception(f'Unsupported header level: {level}')

        return ReportBlock(
            id=ReportBlock.generate_random_id(),
            type=header_type,
            data={'level': level, 'text': content or ''}
        )

    @staticmethod
    def generate_random_id():
        return str(uuid.uuid4())



SHEET_CACHE = {}
IMPORTS = [
    'from kywy.client.kawa_decorators import kawa_tool',
    'from datetime import datetime, date, timedelta',
    'import pandas as pd',
    'import numpy as np',
    'from faker import Faker',
    'import random'
]
LOGS_COLLECTOR = []


def init_logs():
    LOGS_COLLECTOR.clear()


def feedback(application_url, report_file):
    at_least_one_error = bool([m for m in LOGS_COLLECTOR if not m['ok']])
    if at_least_one_error:
        error_messages = '\n'.join([m['message'] for m in LOGS_COLLECTOR if not m['ok']])
        print('❌ There were errors')
        print('#' * 30)
        print('🚨 Please fix these:')
        print('# ERROR SECTION')
        print(error_messages)
        print('# ERROR SECTION')
        print('#' * 30)
        init_logs()
        return error_messages
    else:
        info('---' * 30)
        info(f'🎉 Publication Complete: [{application_url}]')
        info(f'📁 Report file here: [{report_file}]')
        info('---' * 30)
        init_logs()


def info(message: str):
    # TODO: Connect to logger
    LOGS_COLLECTOR.append({'ok': True, 'message': message})
    print(message)


def error(message: str):
    # TODO: Connect to logger
    LOGS_COLLECTOR.append({'ok': False, 'message': message})
    print('🚨' + message)


def start_sync(object_description: str):
    message = f'Now syncing: {object_description}'
    LOGS_COLLECTOR.append({'ok': True, 'message': message})
    print(message)


def to_tuple(i):
    if isinstance(i, str):
        return (i,)
    elif isinstance(i, tuple):
        return i
    elif isinstance(i, list):
        return tuple(i)
    else:
        raise Exception('The variable must be a string, a tuple or a list')


def load_sheet(kawa, sheet_id, skip_cache=False):
    if not sheet_id:
        raise Exception('Please specify a sheet id to load')

    cached_sheet = SHEET_CACHE.get(sheet_id)

    if skip_cache or not cached_sheet:
        SHEET_CACHE[sheet_id] = kawa.entities.sheets().get_entity_by_id(sheet_id)

    return SHEET_CACHE.get(sheet_id)


def load_dashboard(kawa, dashboard_id):
    return kawa.entities.dashboards().get_entity_by_id(dashboard_id)


def _get_column(sheet, column_name):
    sheet_id = sheet['id']
    all_columns = sheet.get('indicatorColumns', []) + sheet.get('computedColumns', [])
    for one_column in all_columns:
        display_name = one_column['displayInformation']['displayName']
        if display_name == column_name:
            return one_column

    raise Exception(f'Column {column_name} not found in sheet {sheet_id}')


def get_column(sheet, column_name, kawa=None, force_refresh_sheet=False):
    try:
        return _get_column(sheet=sheet, column_name=column_name)
    except Exception as e:
        if force_refresh_sheet and kawa:
            # We did not find the column, try to reload the sheet
            refreshed_sheet = load_sheet(
                kawa=kawa,
                sheet_id=sheet['id'],
                skip_cache=True
            )
            return _get_column(sheet=refreshed_sheet, column_name=column_name)
        else:
            raise e


PALETTES = {
    'classic1': ['#8DD3C7', '#FFFFB3', '#BEBADA', '#FB8072', '#80B1D3', '#FDB462', '#B3DE69', '#FCCDE5', '#D9D9D9',
                 '#BC80BD'],
    'classic2': ['#26a1d5', '#ec1254', '#f27c14', '#f5e31d', '#1ee8b6', '#570bb7', '#d042f8', '#2edbef', '#3aefb6',
                 '#f10983'],
    'fire': ["#ff0000", "#ff3000", "#ff6000", "#ff9000", "#ffc000", "#043F98", "#1165C1", "#9CDEEB", "#B7E8EB",
             "#66BEF9"],
    'ice': ["#A8D8E9", "#F0F4F8", "#B2EBF2", "#E0F7FA", "#AEEEEE", "#FFFFFF", "#74C3E8", "#8EDBFF", "#B0E0E6",
            "#007B7F"],
    'constrast': ["#A4D8E1", "#003B5C", "#C2E0E5", "#E7F1F8", "#A9D6E5", "#007B8A", "#B0E1E7", "#F0FFFF", "#C8D6D4",
                  "#2A272E"],
    'autumn': ["#57291F", "#C0413B", "#D77B5F", "#FF9200", "#FFCD73", "#80003A", "#506432", "#FFC500", "#B30019",
               "#EC410B"],
    'tableau': ["#5778A4", "#E49444", "#D1615D", "#85B6B2", "#6A9F58", "#E7CA60", "#A87C9F", "#F1A2A9", "#967662",
                "#B8B0AC"],
    'excel': ["#000000", "#FFFFFF", "#FF0000", "#00FF00", "#0000FF", "#FFFF00", "#FF00FF", "#00FFFF", "#800000",
              "#008000"]
}
